from __future__ import annotations

from operator import methodcaller

import pytest

import ibis
import ibis.common.exceptions as com
from ibis.common.deferred import _


@pytest.mark.parametrize(
    "method",
    [
        methodcaller("tumble", size=ibis.interval(minutes=15)),
        methodcaller(
            "hop", size=ibis.interval(minutes=15), slide=ibis.interval(minutes=1)
        ),
    ],
    ids=["tumble", "hop"],
)
@pytest.mark.parametrize("by", ["g", _.g, ["g"]])
def test_window_by_agg_schema(table, method, by):
    expr = method(table.window_by(time_col=table.i))
    expr = expr.agg(by=by, a_sum=_.a.sum())
    expected_schema = ibis.schema(
        {
            "window_start": "timestamp",
            "window_end": "timestamp",
            "g": "string",
            "a_sum": "int64",
        }
    )
    assert expr.schema() == expected_schema


def test_window_by_with_non_timestamp_column(table):
    with pytest.raises(com.IbisInputError):
        table.window_by(time_col=table.a)


@pytest.mark.parametrize(
    "method",
    [
        methodcaller("tumble", size=ibis.interval(minutes=15)),
        methodcaller(
            "hop", size=ibis.interval(minutes=15), slide=ibis.interval(minutes=1)
        ),
    ],
    ids=["tumble", "hop"],
)
@pytest.mark.parametrize("by", ["g", _.g, ["g"]])
def test_window_by_grouped_agg(table, method, by):
    expr = method(table.window_by(time_col=table.i))
    expr = expr.group_by(by).agg(a_sum=_.a.sum())
    expected_schema = ibis.schema(
        {
            "window_start": "timestamp",
            "window_end": "timestamp",
            "g": "string",
            "a_sum": "int64",
        }
    )
    assert expr.schema() == expected_schema


@pytest.mark.parametrize(
    "method",
    [
        methodcaller("tumble", size=ibis.interval(minutes=15)),
        methodcaller(
            "hop", size=ibis.interval(minutes=15), slide=ibis.interval(minutes=1)
        ),
    ],
    ids=["tumble", "hop"],
)
def test_window_by_global_agg(table, method):
    expr = method(table.window_by(time_col=table.i))
    expr = expr.agg(a_sum=_.a.sum())
    expected_schema = ibis.schema(
        {
            "window_start": "timestamp",
            "window_end": "timestamp",
            "a_sum": "int64",
        }
    )
    assert expr.schema() == expected_schema



from __future__ import annotations

from public import public

import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
from ibis.common.annotations import attribute
from ibis.common.collections import FrozenDict
from ibis.common.patterns import replace
from ibis.common.typing import VarTuple  # noqa: TC001
from ibis.expr.schema import Schema


@public
class PandasRename(ops.Relation):
    parent: ops.Relation
    mapping: FrozenDict[str, str]

    @classmethod
    def from_prefix(cls, parent, prefix):
        mapping = {k: f"{prefix}_{k}" for k in parent.schema}
        return cls(parent, mapping)

    @attribute
    def values(self):
        return FrozenDict(
            {to: ops.Field(self.parent, from_) for from_, to in self.mapping.items()}
        )

    @attribute
    def schema(self):
        return Schema(
            {self.mapping[name]: dtype for name, dtype in self.parent.schema.items()}
        )


@public
class PandasJoin(ops.Relation):
    left: ops.Relation
    right: ops.Relation
    left_on: VarTuple[ops.Value]
    right_on: VarTuple[ops.Value]
    how: str

    @attribute
    def values(self):
        return FrozenDict({**self.left.values, **self.right.values})

    @attribute
    def schema(self):
        return self.left.schema | self.right.schema


@public
class PandasAsofJoin(PandasJoin):
    left_by: VarTuple[ops.Value]
    right_by: VarTuple[ops.Value]
    operator: type


def split_join_predicates(left, right, predicates, only_equality=True):
    left_on = []
    right_on = []
    for pred in predicates:
        if left not in pred.relations or right not in pred.relations:
            # not a usual join predicate, so apply a trick by placing the
            # predicate to the left side and adding a literal True to the right
            # which the left side must be equal to
            left_on.append(pred)
            right_on.append(ops.Literal(True, dtype=dt.boolean))
        elif isinstance(pred, ops.Binary):
            if only_equality and not isinstance(pred, ops.Equals):
                raise TypeError("Only equality join predicates supported with pandas")
            if left in pred.left.relations and right in pred.right.relations:
                left_on.append(pred.left)
                right_on.append(pred.right)
            elif left in pred.right.relations and right in pred.left.relations:
                left_on.append(pred.right)
                right_on.append(pred.left)
            else:
                raise ValueError("Join predicate does not reference both tables")
        else:
            raise TypeError(f"Unsupported join predicate {pred}")

    return left_on, right_on


@replace(ops.JoinChain)
def rewrite_join(_, **kwargs):
    # TODO(kszucs): JoinTable.index can be used as a prefix
    prefixes = {}
    prefixes[_.first] = prefix = str(len(prefixes))
    left = PandasRename.from_prefix(_.first, prefix)

    for link in _.rest:
        prefixes[link.table] = prefix = str(len(prefixes))
        right = PandasRename.from_prefix(link.table, prefix)

        subs = {v: ops.Field(left, k) for k, v in left.values.items()}
        subs.update({v: ops.Field(right, k) for k, v in right.values.items()})
        preds = [pred.replace(subs, filter=ops.Value) for pred in link.predicates]

        # separate ASOF from the rest of the joins
        if link.how == "asof":
            on, *by = preds
            left_on, right_on = split_join_predicates(
                left, right, [on], only_equality=False
            )
            left_by, right_by = split_join_predicates(left, right, by)
            left = PandasAsofJoin(
                how="asof",
                left=left,
                right=right,
                left_on=left_on,
                right_on=right_on,
                left_by=left_by,
                right_by=right_by,
                operator=type(on),
            )
        else:
            # need to replace the fields in the predicates
            left_on, right_on = split_join_predicates(left, right, preds)
            left = PandasJoin(
                how=link.how,
                left=left,
                right=right,
                left_on=left_on,
                right_on=right_on,
            )

    subs = {v: ops.Field(left, k) for k, v in left.values.items()}
    fields = {k: v.replace(subs, filter=ops.Value) for k, v in _.values.items()}
    return ops.Project(left, fields)


@replace(ops.UnboundTable)
def bind_unbound_table(_, backend, **kwargs):
    return ops.DatabaseTable(name=_.name, schema=_.schema, source=backend)



from __future__ import annotations

import hypothesis as h
import hypothesis.strategies as st
import pytest
from packaging.version import parse as vparse

import ibis.expr.datatypes as dt
import ibis.tests.strategies as ibst

np = pytest.importorskip("numpy")
npst = pytest.importorskip("hypothesis.extra.numpy")

from ibis.formats.numpy import NumpySchema, NumpyType  # noqa: E402

roundtripable_types = st.deferred(
    lambda: (
        npst.integer_dtypes(endianness="=")
        | npst.floating_dtypes(endianness="=")
        | npst.datetime64_dtypes(max_period="ns", endianness="=")
        | npst.timedelta64_dtypes(max_period="s", endianness="=")
    )
)


@st.composite
def numpy_schema(draw, item_strategy=roundtripable_types, max_size=10):
    num_fields = draw(st.integers(min_value=0, max_value=max_size))
    names = draw(
        st.lists(st.text(), min_size=num_fields, max_size=num_fields, unique=True)
    )
    types = draw(st.lists(item_strategy, min_size=num_fields, max_size=num_fields))
    return list(zip(names, types))


def assert_dtype_roundtrip(
    numpy_type, ibis_type=None, restored_type=None, nullable=True
):
    dtype = NumpyType.to_ibis(numpy_type, nullable=nullable)
    if ibis_type is not None:
        assert dtype == ibis_type

    nptyp = NumpyType.from_ibis(dtype)
    if restored_type is None:
        restored_type = numpy_type
    assert nptyp == restored_type


@h.given(roundtripable_types)
def test_roundtripable_types(numpy_type):
    assert_dtype_roundtrip(numpy_type, nullable=False)
    assert_dtype_roundtrip(numpy_type, nullable=True)


@h.given(npst.unicode_string_dtypes(endianness="="))
def test_non_roundtripable_str_type(numpy_type):
    assert_dtype_roundtrip(
        numpy_type, dt.String(nullable=False), np.dtype("object"), nullable=False
    )
    assert_dtype_roundtrip(
        numpy_type, dt.String(nullable=True), np.dtype("object"), nullable=True
    )


@h.given(npst.byte_string_dtypes(endianness="="))
def test_non_roundtripable_bytes_type(numpy_type):
    assert_dtype_roundtrip(
        numpy_type, dt.Binary(nullable=False), np.dtype("object"), nullable=False
    )
    assert_dtype_roundtrip(
        numpy_type, dt.Binary(nullable=True), np.dtype("object"), nullable=True
    )


@h.given(
    ibst.null_dtype
    | ibst.variadic_dtypes()
    | ibst.decimal_dtypes()
    | ibst.struct_dtypes()
)
def test_variadic_to_numpy(ibis_type):
    assert NumpyType.from_ibis(ibis_type) == np.dtype("object")


@h.given(ibst.timestamp_dtype())
def test_timestamp_to_numpy(ibis_type):
    assert NumpyType.from_ibis(ibis_type) == np.dtype("datetime64[ns]")


@h.given(ibst.date_dtype())
def test_date_to_numpy(ibis_type):
    assert NumpyType.from_ibis(ibis_type) == np.dtype("datetime64[D]")


@h.given(ibst.time_dtype())
def test_time_to_numpy(ibis_type):
    assert NumpyType.from_ibis(ibis_type) == np.dtype("timedelta64[ns]")


@h.given(ibst.schema())
def test_schema_to_numpy(ibis_schema):
    numpy_schema = NumpySchema.from_ibis(ibis_schema)
    assert len(numpy_schema) == len(ibis_schema)

    for name, numpy_type in numpy_schema:
        assert numpy_type == NumpyType.from_ibis(ibis_schema[name])


@h.given(numpy_schema())
def test_schema_from_numpy(numpy_schema):
    ibis_schema = NumpySchema.to_ibis(numpy_schema)
    assert len(numpy_schema) == len(ibis_schema)

    for name, numpy_type in numpy_schema:
        assert NumpyType.from_ibis(ibis_schema[name]) == numpy_type


@pytest.mark.parametrize(
    ("numpy_dtype", "ibis_dtype"),
    [
        (np.bool_, dt.boolean),
        (np.int8, dt.int8),
        (np.int16, dt.int16),
        (np.int32, dt.int32),
        (np.int64, dt.int64),
        (np.uint8, dt.uint8),
        (np.uint16, dt.uint16),
        (np.uint32, dt.uint32),
        (np.uint64, dt.uint64),
        (np.float16, dt.float16),
        (np.float32, dt.float32),
        (np.float64, dt.float64),
        (np.double, dt.double),
        (np.str_, dt.string),
        (np.datetime64, dt.timestamp),
    ],
)
def test_dtype_from_numpy(numpy_dtype, ibis_dtype):
    assert NumpyType.to_ibis(np.dtype(numpy_dtype)) == ibis_dtype


def test_dtype_from_numpy_dtype_timedelta():
    if vparse(pytest.importorskip("pyarrow").__version__) < vparse("9"):
        pytest.skip("pyarrow < 9 globally mutates the timedelta64 numpy dtype")

    assert NumpyType.to_ibis(np.dtype(np.timedelta64)) == dt.Interval(unit="s")



from __future__ import annotations

from ibis.formats.pandas import PandasData


class BigQueryPandasData(PandasData):
    @classmethod
    def convert_GeoSpatial(cls, s, dtype, pandas_type):
        import geopandas as gpd
        import shapely as shp

        try:
            return gpd.GeoSeries.from_wkt(s)
        except shp.geos.GEOSException:
            return gpd.GeoSeries.from_wkb(s)



from __future__ import annotations

from collections.abc import Iterable, Mapping
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal

import polars as pl

import ibis
import ibis.common.exceptions as com
import ibis.expr.operations as ops
import ibis.expr.schema as sch
import ibis.expr.types as ir
from ibis.backends import BaseBackend, NoUrl
from ibis.backends.polars.compiler import translate
from ibis.backends.polars.rewrites import bind_unbound_table, rewrite_join
from ibis.backends.sql.dialects import Polars
from ibis.common.dispatch import lazy_singledispatch
from ibis.expr.rewrites import lower_stringslice, replace_parameter
from ibis.formats.polars import PolarsSchema
from ibis.util import gen_name, normalize_filename, normalize_filenames

if TYPE_CHECKING:
    from collections.abc import Iterable

    import pandas as pd
    import pyarrow as pa


class Backend(BaseBackend, NoUrl):
    name = "polars"
    dialect = Polars

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._tables = dict()
        self._context = pl.SQLContext()

    def do_connect(
        self, tables: Mapping[str, pl.LazyFrame | pl.DataFrame] | None = None
    ) -> None:
        """Construct a client from a dictionary of polars `LazyFrame`s and/or `DataFrame`s.

        Parameters
        ----------
        tables
            An optional mapping of string table names to polars LazyFrames.

        Examples
        --------
        >>> import ibis
        >>> import polars as pl
        >>> ibis.options.interactive = True
        >>> lazy_frame = pl.LazyFrame(
        ...     {"name": ["Jimmy", "Keith"], "band": ["Led Zeppelin", "Stones"]}
        ... )
        >>> con = ibis.polars.connect(tables={"band_members": lazy_frame})
        >>> t = con.table("band_members")
        >>> t
        ┏━━━━━━━━┳━━━━━━━━━━━━━━┓
        ┃ name   ┃ band         ┃
        ┡━━━━━━━━╇━━━━━━━━━━━━━━┩
        │ string │ string       │
        ├────────┼──────────────┤
        │ Jimmy  │ Led Zeppelin │
        │ Keith  │ Stones       │
        └────────┴──────────────┘
        """
        if tables is not None and not isinstance(tables, Mapping):
            raise TypeError("Input to ibis.polars.connect must be a mapping")

        # tables are emphemeral
        self._tables.clear()

        for name, table in (tables or {}).items():
            self._add_table(name, table)

    def disconnect(self) -> None:
        pass

    @property
    def version(self) -> str:
        return pl.__version__

    def list_tables(self, like=None, database=None):
        return self._filter_with_like(list(self._tables.keys()), like)

    def table(self, name: str, database: None = None) -> ir.Table:
        if database is not None:
            raise com.IbisError(
                "Passing `database` to the Polars backend's `table()` method is not "
                "supported: Polars cannot set a database."
            )

        table = self._tables.get(name)
        if table is None:
            raise com.TableNotFound(name)

        schema = sch.infer(table)
        return ops.DatabaseTable(name, schema, self).to_expr()

    def _register_in_memory_table(self, op: ops.InMemoryTable) -> None:
        self._add_table(op.name, op.data.to_polars(op.schema).lazy())

    def _finalize_memtable(self, name: str) -> None:
        self.drop_table(name, force=True)

    def _add_table(self, name: str, obj: pl.LazyFrame | pl.DataFrame) -> None:
        if isinstance(obj, pl.DataFrame):
            obj = obj.lazy()
        self._tables[name] = obj
        self._context.register(name, obj)

    def sql(
        self, query: str, schema: sch.Schema | None = None, dialect: str | None = None
    ) -> ir.Table:
        query = self._transpile_sql(query, dialect=dialect)
        if schema is None:
            schema = self._get_schema_using_query(query)
        name = ibis.util.gen_name("polars_dot_sql_table")
        self._add_table(name, self._context.execute(query))
        return self.table(name)

    def read_csv(
        self,
        path: str | Path | list[str | Path] | tuple[str | Path],
        table_name: str | None = None,
        **kwargs: Any,
    ) -> ir.Table:
        """Register a CSV file as a table.

        Parameters
        ----------
        path
            The data source. A string or Path to the CSV file.
        table_name
            An optional name to use for the created table. This defaults to
            a sequentially generated name.
        **kwargs
            Additional keyword arguments passed to Polars loading function.
            See https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.scan_csv.html
            for more information.

        Returns
        -------
        ir.Table
            The just-registered table

        """
        source_list = normalize_filenames(path)
        # Flatten the list if there's only one element because Polars
        # can't handle glob strings, or compressed CSVs in a single-element list
        if len(source_list) == 1:
            source_list = source_list[0]
        table_name = table_name or gen_name("read_csv")
        try:
            table = pl.scan_csv(source_list, **kwargs)
            # triggers a schema computation to handle compressed csv inference
            # and raise a compute error
            table.collect_schema()
        except pl.exceptions.ComputeError:
            # handles compressed csvs
            table = pl.read_csv(source_list, **kwargs)

        self._add_table(table_name, table)
        return self.table(table_name)

    def read_json(
        self, path: str | Path, table_name: str | None = None, **kwargs: Any
    ) -> ir.Table:
        """Register a JSON file as a table.

        Parameters
        ----------
        path
            A string or Path to a JSON file; globs are supported
        table_name
            An optional name to use for the created table. This defaults to
            a sequentially generated name.
        **kwargs
            Additional keyword arguments passed to Polars loading function.
            See https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.scan_ndjson.html
            for more information.

        Returns
        -------
        ir.Table
            The just-registered table

        """
        path = normalize_filename(path)
        table_name = table_name or gen_name("read_json")
        try:
            self._add_table(table_name, pl.scan_ndjson(path, **kwargs))
        except pl.exceptions.ComputeError:
            # handles compressed json files
            self._add_table(table_name, pl.read_ndjson(path, **kwargs))
        return self.table(table_name)

    def read_delta(
        self, path: str | Path, table_name: str | None = None, **kwargs: Any
    ) -> ir.Table:
        """Register a Delta Lake as a table in the current database.

        Parameters
        ----------
        path
            The data source(s). Path to a Delta Lake table directory.
        table_name
            An optional name to use for the created table. This defaults to
            a sequentially generated name.
        **kwargs
            Additional keyword arguments passed to Polars loading function.
            See https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.scan_delta.html
            for more information.

        Returns
        -------
        ir.Table
            The just-registered table

        """
        try:
            import deltalake  # noqa: F401
        except ImportError:
            raise ImportError(
                "The deltalake extra is required to use the "
                "read_delta method. You can install it using pip:\n\n"
                "pip install 'ibis-framework[polars,deltalake]'\n"
            )
        path = normalize_filename(path)
        table_name = table_name or gen_name("read_delta")
        self._add_table(table_name, pl.scan_delta(path, **kwargs))
        return self.table(table_name)

    def read_pandas(
        self, source: pd.DataFrame, table_name: str | None = None, **kwargs: Any
    ) -> ir.Table:
        """Register a Pandas DataFrame or pyarrow Table a table in the current database.

        Parameters
        ----------
        source
            The data source.
        table_name
            An optional name to use for the created table. This defaults to
            a sequentially generated name.
        **kwargs
            Additional keyword arguments passed to Polars loading function.
            See https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.from_pandas.html
            for more information.

        Returns
        -------
        ir.Table
            The just-registered table

        """
        table_name = table_name or gen_name("read_in_memory")

        self._add_table(table_name, pl.from_pandas(source, **kwargs).lazy())
        return self.table(table_name)

    def read_parquet(
        self,
        path: str | Path | Iterable[str],
        table_name: str | None = None,
        **kwargs: Any,
    ) -> ir.Table:
        """Register a parquet file as a table in the current database.

        Parameters
        ----------
        path
            The data source(s). May be a path to a file, an iterable of files,
            or directory of parquet files.
        table_name
            An optional name to use for the created table. This defaults to
            a sequentially generated name.
        **kwargs
            Additional keyword arguments passed to Polars loading function.
            See https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.scan_parquet.html
            for more information (if loading a single file or glob; when loading
            multiple files polars' `scan_pyarrow_dataset` method is used instead).

        Returns
        -------
        ir.Table
            The just-registered table

        """
        table_name = table_name or gen_name("read_parquet")
        if not isinstance(path, (str, Path)) and len(path) == 1:
            path = path[0]

        if not isinstance(path, (str, Path)) and len(path) > 1:
            self._import_pyarrow()
            import pyarrow.dataset as ds

            paths = [normalize_filename(p) for p in path]
            obj = pl.scan_pyarrow_dataset(
                source=ds.dataset(paths, format="parquet"),
                **kwargs,
            )
            self._add_table(table_name, obj)
        else:
            path = normalize_filename(path)
            self._add_table(table_name, pl.scan_parquet(path, **kwargs))

        return self.table(table_name)

    def create_table(
        self,
        name: str,
        obj: ir.Table
        | pd.DataFrame
        | pa.Table
        | pa.RecordBatchReader
        | pa.RecordBatch
        | pl.DataFrame
        | pl.LazyFrame
        | None = None,
        *,
        schema: sch.SchemaLike | None = None,
        database: str | None = None,
        temp: bool | None = None,
        overwrite: bool = False,
    ) -> ir.Table:
        if database is not None:
            raise com.IbisError(
                "Passing `database` to the Polars backend's `create_table()` method is "
                "not supported: Polars cannot set a database."
            )

        if temp is False:
            raise com.IbisError(
                "Passing `temp=False` to the Polars backend's `create_table()` method "
                "is not supported: all tables are in memory and temporary."
            )

        if not overwrite and name in self._tables:
            raise com.IntegrityError(
                f"Table {name!r} already exists. Use `overwrite=True` to clobber "
                "existing tables."
            )

        if schema is not None and obj is None:
            obj = pl.LazyFrame([], schema=PolarsSchema.from_ibis(schema))
            self._add_table(name, obj)
        else:
            _read_in_memory(obj, name, self)

        return self.table(name)

    def create_view(
        self,
        name: str,
        obj: ir.Table,
        *,
        database: str | None = None,
        overwrite: bool = False,
    ) -> ir.Table:
        return self.create_table(
            name, obj=obj, temp=None, database=database, overwrite=overwrite
        )

    def drop_table(self, name: str, *, force: bool = False) -> None:
        if name in self._tables:
            del self._tables[name]
            self._context.unregister(name)
        elif not force:
            raise com.IbisError(f"Table {name!r} does not exist")

    def drop_view(self, name: str, *, force: bool = False) -> None:
        self.drop_table(name, force=force)

    def get_schema(self, table_name):
        return self._tables[table_name].schema

    @classmethod
    @lru_cache
    def _get_operations(cls):
        return tuple(op for op in translate.registry if issubclass(op, ops.Value))

    @classmethod
    def has_operation(cls, operation: type[ops.Value]) -> bool:
        # Polars doesn't support geospatial ops, but the dispatcher implements
        # a common base class that makes it appear that it does. Explicitly
        # exclude these operations.
        if issubclass(
            operation, (ops.GeoSpatialUnOp, ops.GeoSpatialBinOp, ops.GeoUnaryUnion)
        ):
            return False
        op_classes = cls._get_operations()
        return operation in op_classes or issubclass(operation, op_classes)

    def compile(
        self, expr: ir.Expr, params: Mapping[ir.Expr, object] | None = None, **_: Any
    ):
        if params is None:
            params = dict()
        else:
            params = {param.op(): value for param, value in params.items()}

        node = expr.as_table().op()
        node = node.replace(
            rewrite_join | replace_parameter | bind_unbound_table | lower_stringslice,
            context={"params": params, "backend": self},
        )

        return translate(node, ctx=self._context)

    def _get_sql_string_view_schema(
        self, *, name: str, table: ir.Table, query: str
    ) -> sch.Schema:
        from ibis.backends.sql.compilers.postgres import compiler

        sql = compiler.add_query_to_expr(name=name, table=table, query=query)
        return self._get_schema_using_query(sql)

    def _get_schema_using_query(self, query: str) -> sch.Schema:
        lazy_frame = self._context.execute(query, eager=False)
        return sch.infer(lazy_frame)

    def _to_dataframe(
        self,
        expr: ir.Expr,
        params: Mapping[ir.Expr, object] | None = None,
        limit: int | None = None,
        streaming: bool = False,
        engine: Literal["cpu", "gpu"] | pl.GPUEngine = "cpu",
        **kwargs: Any,
    ) -> pl.DataFrame:
        self._run_pre_execute_hooks(expr)
        table_expr = expr.as_table()
        lf = self.compile(table_expr, params=params, **kwargs)
        if limit == "default":
            limit = ibis.options.sql.default_limit
        if limit is not None:
            lf = lf.limit(limit)
        df = lf.collect(streaming=streaming, engine=engine)
        # XXX: Polars sometimes returns data with the incorrect column names.
        # For now we catch this case and rename them here if needed.
        expected_cols = tuple(table_expr.columns)
        if tuple(df.columns) != expected_cols:
            df = df.rename(dict(zip(df.columns, expected_cols)))
        return df

    def execute(
        self,
        expr: ir.Expr,
        params: Mapping[ir.Expr, object] | None = None,
        limit: int | None = None,
        streaming: bool = False,
        engine: Literal["cpu", "gpu"] | pl.GPUEngine = "cpu",
        **kwargs: Any,
    ):
        df = self._to_dataframe(
            expr,
            params=params,
            limit=limit,
            streaming=streaming,
            engine=engine,
            **kwargs,
        )
        if isinstance(expr, (ir.Table, ir.Scalar)):
            return expr.__pandas_result__(df.to_pandas())
        else:
            assert isinstance(expr, ir.Column), type(expr)

            dtype = expr.type()
            if dtype.is_temporal():
                return expr.__pandas_result__(df.to_pandas())
            else:
                from ibis.formats.pandas import PandasData

                # note: skip frame-construction overhead
                return PandasData.convert_column(df.to_series().to_pandas(), dtype)

    def to_polars(
        self,
        expr: ir.Expr,
        params: Mapping[ir.Expr, object] | None = None,
        limit: int | None = None,
        streaming: bool = False,
        engine: Literal["cpu", "gpu"] | pl.GPUEngine = "cpu",
        **kwargs: Any,
    ):
        df = self._to_dataframe(
            expr,
            params=params,
            limit=limit,
            streaming=streaming,
            engine=engine,
            **kwargs,
        )
        return expr.__polars_result__(df)

    def _to_pyarrow_table(
        self,
        expr: ir.Expr,
        params: Mapping[ir.Expr, object] | None = None,
        limit: int | None = None,
        streaming: bool = False,
        engine: Literal["cpu", "gpu"] | pl.GPUEngine = "cpu",
        **kwargs: Any,
    ):
        from ibis.formats.pyarrow import PyArrowData

        df = self._to_dataframe(
            expr,
            params=params,
            limit=limit,
            streaming=streaming,
            engine=engine,
            **kwargs,
        )
        return PyArrowData.convert_table(df.to_arrow(), expr.as_table().schema())

    def to_pyarrow(
        self,
        expr: ir.Expr,
        params: Mapping[ir.Expr, object] | None = None,
        limit: int | None = None,
        **kwargs: Any,
    ):
        result = self._to_pyarrow_table(expr, params=params, limit=limit, **kwargs)
        return expr.__pyarrow_result__(result)

    def to_pyarrow_batches(
        self,
        expr: ir.Expr,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        chunk_size: int = 1_000_000,
        **kwargs: Any,
    ):
        self._import_pyarrow()
        table = self._to_pyarrow_table(expr, params=params, limit=limit, **kwargs)
        return table.to_reader(chunk_size)

    def _create_cached_table(self, name, expr):
        return self.create_table(name, self.compile(expr).cache())

    def _drop_cached_table(self, name):
        self.drop_table(name, force=True)


@lazy_singledispatch
def _read_in_memory(source: Any, table_name: str, _conn: Backend, **kwargs: Any):
    raise NotImplementedError(
        f"The `{_conn.name}` backend currently does not support "
        f"reading data of {type(source)!r}"
    )


@_read_in_memory.register("ibis.expr.types.Table")
def _table(source, table_name, _conn, **kwargs: Any):
    _conn._add_table(table_name, source.to_polars())


@_read_in_memory.register("polars.DataFrame")
@_read_in_memory.register("polars.LazyFrame")
def _polars(source, table_name, _conn, **kwargs: Any):
    _conn._add_table(table_name, source)


@_read_in_memory.register("pyarrow.Table")
@_read_in_memory.register("pyarrow.RecordBatchReader")
@_read_in_memory.register("pyarrow.RecordBatch")
def _pyarrow(source, table_name, _conn, **kwargs: Any):
    _conn._add_table(table_name, pl.from_arrow(source, **kwargs).lazy())


@_read_in_memory.register("pandas.DataFrame")
def _pandas(source: pd.DataFrame, table_name, _conn, **kwargs: Any):
    _conn._add_table(table_name, pl.from_pandas(source, **kwargs).lazy())



from __future__ import annotations

import pytest
from pyflink.common.types import Row

import ibis
from ibis.backends.tests.errors import Py4JJavaError


@pytest.mark.parametrize(
    ("data", "schema", "expected"),
    [
        pytest.param(
            {"value": [{"a": 1}, {"a": 2}]},
            {"value": "!struct<a: !int>"},
            [Row(Row([1])), Row(Row([2]))],
            id="simple_named_struct",
        ),
        pytest.param(
            {"value": [[{"a": 1}, {"a": 2}], [{"a": 3}, {"a": 4}]]},
            {"value": "!array<!struct<a: !int>>"},
            [Row([Row([1]), Row([2])]), Row([Row([3]), Row([4])])],
            id="single_field_named_struct_array",
        ),
        pytest.param(
            {"value": [[{"a": 1, "b": 2}, {"a": 2, "b": 2}]]},
            {"value": "!array<!struct<a: !int, b: !int>>"},
            [Row([Row([1, 2]), Row([2, 2])])],
            id="named_struct_array",
        ),
    ],
)
def test_create_memtable(con, data, schema, expected):
    t = ibis.memtable(data, schema=ibis.schema(schema))
    # cannot use con.execute(t) directly because of some behavioral discrepancy between
    # `TableEnvironment.execute_sql()` and `TableEnvironment.sql_query()`; this doesn't
    # seem to be an issue if we don't execute memtable directly
    result = list(con.raw_sql(con.compile(t)).collect())
    for element in expected:
        assert element in result


@pytest.mark.notyet(
    ["flink"],
    raises=Py4JJavaError,
    reason="cannot create an ARRAY of named STRUCTs directly from the ARRAY[] constructor; https://issues.apache.org/jira/browse/FLINK-34898",
)
def test_create_named_struct_array_with_array_constructor(con):
    con.raw_sql("SELECT ARRAY[cast(ROW(1) as ROW<a INT>)];")



from __future__ import annotations

import collections
import decimal

from ibis.common.dispatch import lazy_singledispatch


def test_lazy_singledispatch():
    @lazy_singledispatch
    def foo(x):
        """A docstring."""
        return "base result"

    @foo.register(int)
    def _(x):
        return x + 1

    @foo.register(float)
    def _(x):
        return x - 1

    @foo.register((tuple, list))
    def _(x):
        return tuple(foo(i) for i in x)

    class Bar:
        pass

    assert foo.__name__ == "foo"
    assert foo.__doc__ == "A docstring."

    assert foo(1) == 2
    assert foo(1.0) == 0.0
    assert foo((1, 2.0)) == (2, 1.0)
    assert foo([1, 2.0]) == (2, 1.0)
    assert foo("a string") == "base result"


def test_lazy_singledispatch_extra_args():
    @lazy_singledispatch
    def foo(*args, **kwargs):
        pass

    @foo.register(int)
    def _(a, b, c=2):
        return a + b + c

    @foo.register(float)
    def _(a, b, c=2):
        return a - b - c

    assert foo(1, 2) == 5
    assert foo(1.0, 2) == -3.0
    assert foo(1, 2, c=3) == 6


def test_lazy_singledispatch_lazy():
    @lazy_singledispatch
    def foo(a):
        return a

    @foo.register("decimal.Decimal")
    def inc(a):
        return a + 1

    assert foo(1) == 1
    assert foo(decimal.Decimal(1)) == decimal.Decimal(2)


def test_lazy_singledispatch_lazy_walks_mro():
    class Subclass(decimal.Decimal):
        pass

    class Subclass2(decimal.Decimal):
        pass

    @lazy_singledispatch
    def foo(a):
        return "base call"

    @foo.register(Subclass2)
    def _(a):
        return "eager call"

    @foo.register("decimal.Decimal")
    def _(a):
        return "lazy call"

    assert foo(1) == "base call"
    assert foo(Subclass2(1)) == "eager call"
    assert foo(Subclass(1)) == "lazy call"
    # not overwritten by lazy loader
    assert foo(Subclass2(1)) == "eager call"


def test_lazy_singledispatch_abc():
    class mydict(dict):
        pass

    @lazy_singledispatch
    def foo(a):
        return "base"

    @foo.register(collections.abc.Mapping)
    def _(a):
        return "mapping"

    @foo.register(mydict)
    def _(a):
        return "mydict"

    @foo.register(collections.abc.Callable)
    def _(a):
        return "callable"

    assert foo(1) == "base"
    assert foo({}) == "mapping"
    assert foo(mydict()) == "mydict"  # concrete takes precedence
    assert foo(sum) == "callable"



import ibis


functional_alltypes = ibis.table(
    name="functional_alltypes",
    schema={
        "id": "int32",
        "bool_col": "boolean",
        "tinyint_col": "int8",
        "smallint_col": "int16",
        "int_col": "int32",
        "bigint_col": "int64",
        "float_col": "float32",
        "double_col": "float64",
        "date_string_col": "string",
        "string_col": "string",
        "timestamp_col": "timestamp",
        "year": "int32",
        "month": "int32",
    },
)

result = functional_alltypes.filter(
    (
        functional_alltypes.timestamp_col
        < ((ibis.timestamp("2010-01-01 00:00:00") + ibis.interval(3)))
    ),
    (functional_alltypes.timestamp_col < ((ibis.now() + ibis.interval(10)))),
).count()



from __future__ import annotations

import re

import pytest
from pytest import param

import ibis
import ibis.common.exceptions as exc
from ibis import _
from ibis.backends.conftest import _get_backends_to_test

sg = pytest.importorskip("sqlglot")


@pytest.mark.parametrize(
    "expr",
    [
        param(ibis.literal(432), id="simple_literal"),
        param(
            ibis.array([432]),
            marks=[
                pytest.mark.never(
                    ["mysql", "mssql", "oracle", "impala", "sqlite"],
                    raises=(exc.OperationNotDefinedError, exc.UnsupportedBackendType),
                    reason="arrays not supported in the backend",
                ),
            ],
            id="array_literal",
        ),
        param(
            ibis.struct(dict(abc=432)),
            marks=[
                pytest.mark.never(
                    ["impala", "mysql", "sqlite", "mssql", "exasol"],
                    raises=(NotImplementedError, exc.UnsupportedBackendType),
                    reason="structs not supported in the backend",
                ),
                pytest.mark.notimpl(
                    ["mssql"], reason="struct literals are not yet implemented"
                ),
            ],
            id="struct_literal",
        ),
    ],
)
@pytest.mark.never(["polars"], reason="not SQL", raises=ValueError)
def test_literal(backend, expr):
    assert "432" in ibis.to_sql(expr, dialect=backend.name())


@pytest.mark.never(["polars"], reason="not SQL", raises=ValueError)
def test_group_by_has_index(backend, snapshot):
    countries = ibis.table(
        dict(continent="string", population="int64"), name="countries"
    )
    expr = countries.group_by(
        cont=(
            _.continent.cases(
                ("NA", "North America"),
                ("SA", "South America"),
                ("EU", "Europe"),
                ("AF", "Africa"),
                ("AS", "Asia"),
                ("OC", "Oceania"),
                ("AN", "Antarctica"),
                else_="Unknown continent",
            )
        )
    ).agg(total_pop=_.population.sum())
    sql = str(ibis.to_sql(expr, dialect=backend.name()))
    snapshot.assert_match(sql, "out.sql")


@pytest.mark.never(["polars"], reason="not SQL", raises=ValueError)
def test_cte_refs_in_topo_order(backend, snapshot):
    mr0 = ibis.table(schema=ibis.schema(dict(key="int")), name="leaf")

    mr1 = mr0.filter(ibis.literal(True))

    mr2 = mr1.join(mr1[["key"]], ["key"])
    mr3 = mr2.join(mr2, ["key"])

    sql = str(ibis.to_sql(mr3, dialect=backend.name()))
    snapshot.assert_match(sql, "out.sql")


@pytest.mark.never(["polars"], reason="not SQL", raises=ValueError)
def test_isin_bug(con, snapshot):
    t = ibis.table(dict(x="int"), name="t")
    good = t.filter(t.x > 2).x
    expr = t.x.isin(good)
    snapshot.assert_match(str(ibis.to_sql(expr, dialect=con.name)), "out.sql")


@pytest.mark.never(["polars"], reason="not SQL", raises=ValueError)
@pytest.mark.notyet(
    ["exasol", "oracle", "flink"],
    reason="no unnest support",
    raises=exc.OperationNotDefinedError,
)
@pytest.mark.notyet(
    ["risingwave"], reason="no arbitrary support", raises=exc.OperationNotDefinedError
)
@pytest.mark.notyet(
    ["sqlite", "mysql", "druid", "impala", "mssql"], reason="no unnest support upstream"
)
@pytest.mark.parametrize("backend_name", _get_backends_to_test())
def test_union_aliasing(backend_name, snapshot):
    if backend_name == "snowflake":
        pytest.skip(
            "pivot requires random names to make snowflake work properly and cannot be snapshotted"
        )
    from ibis import selectors as s

    t = ibis.table(
        {
            "field_of_study": "string",
            "1970-71": "int64",
            "1975-76": "int64",
            "1980-81": "int64",
            "1985-86": "int64",
            "1990-91": "int64",
            "1995-96": "int64",
            "2000-01": "int64",
            "2005-06": "int64",
            "2010-11": "int64",
            "2011-12": "int64",
            "2012-13": "int64",
            "2013-14": "int64",
            "2014-15": "int64",
            "2015-16": "int64",
            "2016-17": "int64",
            "2017-18": "int64",
            "2018-19": "int64",
            "2019-20": "int64",
        },
        name="humanities",
    )

    years_window = ibis.window(order_by=_.years, group_by=_.field_of_study)
    diff_agg = (
        t.pivot_longer(s.matches(r"\d{4}-\d{2}"), names_to="years", values_to="degrees")
        .mutate(
            earliest_degrees=_.degrees.first().over(years_window),
            latest_degrees=_.degrees.last().over(years_window),
        )
        .mutate(diff=_.latest_degrees - _.earliest_degrees)
        .group_by(_.field_of_study)
        .agg(diff=_.diff.arbitrary())
    )

    top_ten = diff_agg.order_by(_.diff.desc()).limit(10)
    bottom_ten = diff_agg.filter(_.diff < 0).order_by(_.diff).limit(10)

    result = top_ten.union(bottom_ten)

    snapshot.assert_match(str(ibis.to_sql(result, dialect=backend_name)), "out.sql")


@pytest.mark.never(["polars"], reason="not SQL", raises=ValueError)
@pytest.mark.parametrize(
    "value",
    [
        param(
            ibis.random(),
            marks=pytest.mark.notimpl(
                ["risingwave", "druid"], raises=exc.OperationNotDefinedError
            ),
            id="random",
        ),
        param(
            ibis.uuid(),
            marks=pytest.mark.notimpl(
                ["exasol", "risingwave", "druid", "oracle", "pyspark"],
                raises=exc.OperationNotDefinedError,
            ),
            id="uuid",
        ),
    ],
)
def test_selects_with_impure_operations_not_merged(con, snapshot, value):
    t = ibis.table({"x": "int64", "y": "float64"}, name="t")
    t = t.mutate(y=value, z=value)
    t = t.mutate(size=(t.y == t.z).ifelse("big", "small"))

    sql = str(ibis.to_sql(t, dialect=con.name))
    snapshot.assert_match(sql, "out.sql")


@pytest.mark.never(["polars"], reason="not SQL", raises=NotImplementedError)
def test_to_sql_default_backend(con, snapshot, monkeypatch):
    monkeypatch.setattr(ibis.options, "default_backend", con)

    t = ibis.memtable({"b": [1, 2]}, name="mytable")
    expr = t.select("b").count()
    snapshot.assert_match(ibis.to_sql(expr), "to_sql.sql")


@pytest.mark.notimpl(["polars"], raises=ValueError, reason="not a SQL backend")
def test_many_subqueries(backend_name, snapshot):
    def query(t, group_cols):
        t2 = t.mutate(key=ibis.row_number().over(ibis.window(order_by=group_cols)))
        return t2.inner_join(t2[["key"]], "key")

    t = ibis.table(dict(street="str"), name="data")

    t2 = query(t, group_cols=["street"])
    t3 = query(t2, group_cols=["street"])

    snapshot.assert_match(str(ibis.to_sql(t3, dialect=backend_name)), "out.sql")


@pytest.mark.parametrize("backend_name", _get_backends_to_test())
@pytest.mark.notimpl(["polars"], raises=ValueError, reason="not a SQL backend")
def test_mixed_qualified_and_unqualified_predicates(backend_name, snapshot):
    t = ibis.table({"x": "int64"}, name="t")
    expr = t.mutate(y=t.x.sum().over(ibis.window())).filter(
        _.y <= 37, _.x.mean().over().notnull()
    )
    result = ibis.to_sql(expr, dialect=backend_name)

    sc = ibis.backends.sql.compilers
    compiler = getattr(sc, backend_name).compiler

    assert (not compiler.supports_qualify) or re.search(
        r"\bQUALIFY\b", result, flags=re.MULTILINE | re.IGNORECASE
    )
    snapshot.assert_match(result, "out.sql")


@pytest.mark.parametrize("backend_name", _get_backends_to_test())
@pytest.mark.notimpl(["polars"], raises=ValueError, reason="not a SQL backend")
@pytest.mark.notimpl(
    ["druid", "risingwave"],
    raises=exc.OperationNotDefinedError,
    reason="random not supported",
)
def test_rewrite_context(snapshot, backend_name):
    table = ibis.memtable({"test": [1, 2, 3, 4, 5]}, name="test")
    expr = table.select(new_col=ibis.ntile(2).over(order_by=ibis.random())).limit(10)
    result = ibis.to_sql(expr, dialect=backend_name)
    snapshot.assert_match(result, "out.sql")


@pytest.mark.parametrize("subquery", [False, True], ids=["subquery", "table"])
@pytest.mark.parametrize("backend_name", _get_backends_to_test())
@pytest.mark.notimpl(["polars"], raises=ValueError, reason="not a SQL backend")
@pytest.mark.notimpl(
    ["druid", "risingwave"],
    raises=exc.OperationNotDefinedError,
    reason="sample not supported",
)
def test_sample(backend_name, snapshot, subquery):
    t = ibis.table({"x": "int64", "y": "int64"}, name="test")
    if subquery:
        t = t.filter(t.x > 10)
    block = ibis.to_sql(t.sample(0.5, method="block"), dialect=backend_name)
    row = ibis.to_sql(t.sample(0.5, method="row"), dialect=backend_name)
    snapshot.assert_match(block, "block.sql")
    snapshot.assert_match(row, "row.sql")


@pytest.mark.parametrize("backend_name", _get_backends_to_test())
@pytest.mark.notimpl(["polars"], raises=ValueError, reason="not a SQL backend")
def test_order_by_no_deference_literals(backend_name, snapshot):
    t = ibis.table({"a": "int"}, name="test")
    s = t.select("a", i=ibis.literal(9), s=ibis.literal("foo"))
    o = s.order_by("a", "i", "s")
    sql = ibis.to_sql(o, dialect=backend_name)
    snapshot.assert_match(sql, "out.sql")



from __future__ import annotations

from datetime import timedelta

import pandas as pd
import pandas.testing as tm
import pyspark.sql.types as pt
import pytest
from pytest import param

import ibis
from ibis.common.exceptions import IbisTypeError

pyspark = pytest.importorskip("pyspark")


@pytest.fixture
def t(con):
    return con.table("basic_table")


@pytest.fixture
def df(con):
    return con._session.table("basic_table").toPandas()


def test_basic(t):
    result = t.execute()
    expected = pd.DataFrame({"id": range(10), "str_col": "value"})
    tm.assert_frame_equal(result, expected)


def test_projection(t):
    result1 = t.mutate(v=t["id"]).execute()
    expected1 = pd.DataFrame({"id": range(10), "str_col": "value", "v": range(10)})

    result2 = t.mutate(v=t["id"]).mutate(v2=t["id"]).mutate(id=t["id"] * 2).execute()
    expected2 = pd.DataFrame(
        {
            "id": range(0, 20, 2),
            "str_col": "value",
            "v": range(10),
            "v2": range(10),
        }
    )

    tm.assert_frame_equal(result1, expected1)
    tm.assert_frame_equal(result2, expected2)


def test_aggregation_col(t, df):
    result = t["id"].count().execute()
    assert result == len(df)


def test_aggregation(t, df):
    result = t.aggregate(max=t["id"].max()).execute()
    expected = pd.DataFrame({"max": [df.id.max()]})
    tm.assert_frame_equal(result, expected)


def test_group_by(t, df):
    result = t.group_by("id").aggregate(max=t["id"].max()).execute()
    expected = df[["id"]].assign(max=df.groupby("id").id.max())
    tm.assert_frame_equal(result, expected)


def test_window(t, df):
    w = ibis.window()
    result = t.mutate(grouped_demeaned=t["id"] - t["id"].mean().over(w)).execute()
    expected = df.assign(grouped_demeaned=df.id - df.id.mean())

    tm.assert_frame_equal(result, expected)


def test_greatest(t, df):
    result = t.mutate(greatest=ibis.greatest(t.id, t.id + 1)).execute()
    expected = df.assign(greatest=df.id + 1)
    tm.assert_frame_equal(result, expected)


@pytest.mark.parametrize(
    ("filter_fn", "expected_fn"),
    [
        param(lambda t: t.filter(t.id < 5), lambda df: df[df.id < 5]),
        param(lambda t: t.filter(t.id != 5), lambda df: df[df.id != 5]),
        param(
            lambda t: t.filter([t.id < 5, t.str_col == "na"]),
            lambda df: df[df.id < 5][df.str_col == "na"],
        ),
        param(
            lambda t: t.filter((t.id > 3) & (t.id < 11)),
            lambda df: df[(df.id > 3) & (df.id < 11)],
        ),
        param(
            lambda t: t.filter((t.id == 3) | (t.id == 5)),
            lambda df: df[(df.id == 3) | (df.id == 5)],
        ),
    ],
)
def test_filter(t, df, filter_fn, expected_fn):
    result = filter_fn(t).execute().reset_index(drop=True)
    expected = expected_fn(df).reset_index(drop=True)
    tm.assert_frame_equal(result, expected)


def test_cast(t, df):
    result = t.mutate(id_string=t.id.cast("string")).execute()
    df = df.assign(id_string=df.id.astype(str))
    tm.assert_frame_equal(result, df)


def test_alias_after_select(t, df):
    # Regression test for issue 2136
    table = t[["id"]]
    table = table.mutate(id2=table["id"])
    result = table.execute()
    tm.assert_series_equal(result["id"], result["id2"], check_names=False)


def test_interval_columns_invalid(con):
    df_interval_invalid = con._session.createDataFrame(
        [[timedelta(days=10, hours=10, minutes=10, seconds=10)]],
        pt.StructType(
            [
                pt.StructField(
                    "interval_day_hour",
                    pt.DayTimeIntervalType(
                        pt.DayTimeIntervalType.DAY, pt.DayTimeIntervalType.SECOND
                    ),
                )
            ]
        ),
    )

    df_interval_invalid.createTempView("invalid_interval_table")
    msg = r"DayTimeIntervalType.+ couldn't be converted to Interval"
    with pytest.raises(IbisTypeError, match=msg):
        con.table("invalid_interval_table")


def test_string_literal_backslash_escaping(con):
    expr = ibis.literal("\\d\\e")
    result = con.execute(expr)
    assert result == "\\d\\e"


def test_connect_without_explicit_session():
    con = ibis.pyspark.connect()
    result = con.sql("SELECT CAST(1 AS BIGINT) as foo").to_pandas()
    tm.assert_frame_equal(result, pd.DataFrame({"foo": [1]}))



from __future__ import annotations

import json
import os
from collections import Counter

import pandas as pd
import pandas.testing as tm
import pyarrow as pa
import pytest
from pytest import param

import ibis
import ibis.common.exceptions as com
from ibis.backends.snowflake.tests.conftest import _get_url
from ibis.util import gen_name


@pytest.fixture
def temp_catalog(con):
    cat = gen_name("tmp_catalog")

    con.create_catalog(cat)
    assert cat in con.list_catalogs()

    yield cat

    con.drop_catalog(cat)
    assert cat not in con.list_catalogs()


@pytest.fixture
def temp_db(con, temp_catalog):
    database = gen_name("tmp_database")

    con.create_database(database, catalog=temp_catalog)
    assert database in con.list_databases(catalog=temp_catalog)

    yield database

    con.drop_database(database, catalog=temp_catalog)
    assert database not in con.list_databases(catalog=temp_catalog)


def test_cross_db_access(con, temp_catalog, temp_db):
    table = gen_name("tmp_table")
    con.raw_sql(
        f'CREATE TABLE "{temp_catalog}"."{temp_db}"."{table}" ("x" INT)'
    ).close()
    t = con.table(table, database=(temp_catalog, temp_db))
    assert t.schema() == ibis.schema(dict(x="int"))
    assert t.execute().empty


def test_cross_db_create_table(con, temp_catalog, temp_db):
    table_name = gen_name("tmp_table")
    data = pd.DataFrame({"key": list("abc"), "value": [[1], [2], [3]]})
    table = con.create_table(table_name, data, database=f"{temp_catalog}.{temp_db}")
    queried_table = con.table(table_name, database=(temp_catalog, temp_db))

    tm.assert_frame_equal(table.execute(), data)
    tm.assert_frame_equal(queried_table.execute(), data)


@pytest.fixture(scope="session")
def simple_con():
    return ibis.connect(_get_url())


@pytest.mark.parametrize(
    "data",
    [
        # raw
        {"key": list("abc"), "value": [[1], [2], [3]]},
        # dataframe
        pd.DataFrame({"key": list("abc"), "value": [[1], [2], [3]]}),
        # pyarrow table
        pa.Table.from_pydict({"key": list("abc"), "value": [[1], [2], [3]]}),
    ],
)
def test_basic_memtable_registration(simple_con, data):
    expected = pd.DataFrame({"key": list("abc"), "value": [[1], [2], [3]]})
    t = ibis.memtable(data)
    result = simple_con.execute(t)
    tm.assert_frame_equal(result, expected)


def test_repeated_memtable_registration(simple_con, mocker):
    data = {"key": list("abc"), "value": [[1], [2], [3]]}
    expected = pd.DataFrame(data)
    t = ibis.memtable(data)

    spy = mocker.spy(simple_con, "_register_in_memory_table")

    n = 2

    for _ in range(n):
        tm.assert_frame_equal(simple_con.execute(t), expected)

    # assert that we called _register_in_memory_table exactly once
    spy.assert_called_once()


def test_timestamp_tz_column(simple_con):
    t = simple_con.create_table(
        ibis.util.gen_name("snowflake_timestamp_tz_column"),
        schema=ibis.schema({"ts": "string"}),
        temp=True,
    ).mutate(ts=lambda t: t.ts.as_timestamp("YYYY-MM-DD HH24-MI-SS"))
    expr = t.ts
    assert expr.execute().empty


def test_create_database(simple_con):
    database = gen_name("test_create_database")

    cur_database = simple_con.current_database
    cur_catalog = simple_con.current_catalog

    simple_con.create_database(database)

    assert simple_con.current_database == cur_database
    assert simple_con.current_catalog == cur_catalog

    simple_con.drop_database(database)

    assert simple_con.current_database == cur_database
    assert simple_con.current_catalog == cur_catalog


def test_create_catalog(simple_con):
    catalog = gen_name("test_create_catalog")
    cur_catalog = simple_con.current_catalog

    simple_con.create_catalog(catalog)
    assert simple_con.current_catalog == cur_catalog

    simple_con.drop_catalog(catalog)
    assert simple_con.current_catalog == cur_catalog


@pytest.fixture(scope="session")
def cat_con():
    return ibis.connect(_get_url())


@pytest.fixture(scope="session")
def db_con():
    return ibis.connect(_get_url())


def test_drop_current_catalog_not_allowed(cat_con):
    catalog = gen_name("test_create_catalog")
    cur_cat = cat_con.current_catalog

    cat_con.create_catalog(catalog)

    assert cat_con.current_catalog == cur_cat

    cat_con.raw_sql(f'USE DATABASE "{catalog}"').close()

    with pytest.raises(com.UnsupportedOperationError, match="behavior is undefined"):
        cat_con.drop_catalog(catalog)

    cat_con.raw_sql(f'USE DATABASE "{cur_cat}"').close()

    cat_con.drop_catalog(catalog)


def test_drop_current_db_not_allowed(db_con):
    database = gen_name("test_create_database")
    cur_database = db_con.current_database

    db_con.create_database(database)

    assert db_con.current_database == cur_database

    db_con.raw_sql(f'USE SCHEMA "{database}"').close()

    with pytest.raises(com.UnsupportedOperationError, match="behavior is undefined"):
        db_con.drop_database(database)

    db_con.raw_sql(f'USE SCHEMA "{cur_database}"').close()

    db_con.drop_database(database)


def test_read_csv_options(con, tmp_path):
    path = tmp_path / "test_pipe.csv"
    path.write_text("a|b\n1|2\n3|4\n")

    t = con.read_csv(path, field_delimiter="|")

    assert t.schema() == ibis.schema(dict(a="int64", b="int64"))


def test_read_csv_https(con):
    t = con.read_csv(
        "https://storage.googleapis.com/ibis-tutorial-data/wowah_data/locations.csv",
        field_optionally_enclosed_by='"',
    )
    assert t.schema() == ibis.schema(
        {
            "Map_ID": "int64",
            "Location_Type": "string",
            "Location_Name": "string",
            "Game_Version": "string",
        }
    )
    assert t.count().execute() == 151


@pytest.fixture(scope="module")
def json_data():
    return [
        {"a": 1, "b": "abc", "c": [{"d": 1}]},
        {"a": 2, "b": "def", "c": [{"d": 2}]},
        {"a": 3, "b": "ghi", "c": [{"d": 3}]},
    ]


@pytest.mark.parametrize(
    "serialize",
    [
        param(lambda obj: "\n".join(map(json.dumps, obj)), id="ndjson"),
        param(json.dumps, id="json"),
    ],
)
def test_read_json(con, tmp_path, serialize, json_data):
    path = tmp_path / "test.json"
    path.write_text(serialize(json_data))

    t = con.read_json(path)

    assert t.schema() == ibis.schema(dict(a="int", b="string", c="array<json>"))
    assert t.count().execute() == len(json_data)


def test_read_parquet(con, data_dir):
    path = data_dir / "parquet" / "functional_alltypes.parquet"

    t = con.read_parquet(path)

    assert t.timestamp_col.type().is_timestamp()


def test_array_repr(con, monkeypatch):
    monkeypatch.setattr(ibis.options, "interactive", True)
    t = con.tables.ARRAY_TYPES
    expr = t.x
    assert repr(expr)


def test_insert(con):
    name = gen_name("test_insert")

    t = con.create_table(
        name, schema=ibis.schema({"a": "int", "b": "string", "c": "int"}), temp=True
    )
    assert t.count().execute() == 0

    expected = pd.DataFrame({"a": [1, 2, 3], "b": ["x", "y", None], "c": [2, None, 3]})

    con.insert(name, ibis.memtable(expected))

    result = t.order_by("a").execute()

    tm.assert_frame_equal(result, expected)

    con.insert(name, expected)
    assert t.count().execute() == 6

    con.insert(name, expected, overwrite=True)
    assert t.count().execute() == 3

    name = gen_name("test_insert")

    t = con.create_table(
        name,
        schema=ibis.schema(
            {
                "ID": "int",
                "NAME": "string",
                "SCORE_1": "float",
                "SCORE_2": "float",
                "SCORE_3": "float",
                "AGE": "int",
            }
        ),
        temp=True,
        overwrite=True,
    )
    con.insert(
        name,
        obj=[
            {
                "ID": 10000,
                "NAME": "....",
                "SCORE_1": 0.2,
                "SCORE_2": 0.5,
                "SCORE_3": 0.75,
                "AGE": 1000,
            }
        ],
    )
    assert t.columns == ("ID", "NAME", "SCORE_1", "SCORE_2", "SCORE_3", "AGE")
    assert t.count().execute() == 1


def test_compile_does_not_make_requests(con, mocker):
    astronauts = con.table("astronauts")
    expr = astronauts.year_of_selection.value_counts()
    spy = mocker.spy(con.con, "cursor")
    assert expr.compile() is not None
    spy.assert_not_called()

    t = ibis.memtable({"a": [1, 2, 3]})
    assert con.compile(t) is not None
    spy.assert_not_called()

    assert ibis.to_sql(t, dialect="snowflake") is not None
    spy.assert_not_called()

    assert ibis.to_sql(expr) is not None
    spy.assert_not_called()


# this won't be hit in CI, but folks can test locally
@pytest.mark.xfail(
    condition=os.environ.get("SNOWFLAKE_HOME") is None,
    reason="SNOWFLAKE_HOME is not set",
)
@pytest.mark.xfail(
    condition=os.environ.get("SNOWFLAKE_DEFAULT_CONNECTION_NAME") is None,
    reason="SNOWFLAKE_DEFAULT_CONNECTION_NAME is not set",
)
def test_no_argument_connection():
    con = ibis.snowflake.connect()
    assert con.list_tables() is not None

    con = ibis.connect("snowflake://")
    assert con.list_tables() is not None


def test_struct_of_json(con):
    raw = {"a": [1, 2, 3], "b": "456"}
    lit = ibis.struct(raw)
    expr = lit.cast("struct<a: array<int>, b: json>")

    n = 5
    t = con.tables.functional_alltypes.mutate(lit=expr).limit(n).lit
    result = con.to_pyarrow(t)

    assert len(result) == n
    assert all(value == raw for value in result.to_pylist())


@pytest.mark.parametrize(
    "database",
    ["IBIS_TESTING.INFORMATION_SCHEMA", ("IBIS_TESTING", "INFORMATION_SCHEMA")],
    ids=["dotted-path", "tuple"],
)
def test_list_tables_with_database(con, database):
    like_table = {
        "EVENT_TABLES",
        "EXTERNAL_TABLES",
        "HYBRID_TABLES",
        "TABLES",
        "TABLE_CONSTRAINTS",
        "TABLE_PRIVILEGES",
        "TABLE_STORAGE_METRICS",
    }
    tables = con.list_tables(database=database, like="TABLE")
    assert like_table.issubset(tables)


def test_timestamp_memtable(con):
    df = pd.DataFrame(
        {
            "ts": [
                pd.Timestamp("1970-01-01 00:00:00"),
                pd.Timestamp("1970-01-01 00:00:01"),
                pd.Timestamp("1970-01-01 00:00:02"),
            ]
        }
    )
    t = ibis.memtable(df)
    result = con.to_pandas(t)
    tm.assert_frame_equal(result, df)


def test_connect_without_snowflake_url():
    # We're testing here that the non-URL connection works.
    # Specifically that a `database` location passed in as "catalog/database"
    # will be parsed correctly
    user = os.getenv("SNOWFLAKE_USER")
    account = os.getenv("SNOWFLAKE_ACCOUNT")
    catalog = os.getenv("SNOWFLAKE_DATABASE")
    database = os.getenv("SNOWFLAKE_SCHEMA")
    warehouse = os.getenv("SNOWFLAKE_WAREHOUSE")

    if (password := os.getenv("SNOWFLAKE_PASSWORD")) is None:
        pytest.skip(reason="No snowflake password set, nothing to do here")

    database = "/".join(filter(None, (catalog, database)))

    nonurlcon = ibis.snowflake.connect(
        user=user,
        account=account,
        database=database,
        warehouse=warehouse,
        password=password,
    )

    assert nonurlcon.list_tables()


def test_table_unnest_with_empty_strings(con):
    t = ibis.memtable({"x": [["", ""], [""], [], None]})
    expected = Counter(["", "", "", None, None])
    expr = t.unnest(t.x)["x"]
    result = con.execute(expr)
    assert Counter(result.values) == expected


def test_insert_dict_variants(con):
    name = gen_name("test_insert_dict_variants")

    t = con.create_table(name, schema=ibis.schema({"a": "int", "b": "str"}), temp=True)
    assert len(t.execute()) == 0

    data = [{"a": 1, "b": "a"}, {"a": 2, "b": "b"}]

    con.insert(name, data)
    assert len(t.execute()) == 2

    con.insert(name, ibis.memtable(data))
    assert len(t.execute()) == 4



from __future__ import annotations

import duckdb
import pytest
from pytest import param

import ibis
from ibis import udf


@udf.scalar.builtin
def hamming(a: str, b: str) -> int: ...


@udf.scalar.builtin
def jaccard(a: str, b: str) -> float: ...


@udf.scalar.builtin
def jaro_similarity(a: str, b: str) -> float: ...


@udf.scalar.builtin
def jaro_winkler_similarity(a: str, b: str) -> float: ...


@udf.scalar.builtin
def damerau_levenshtein(a: str, b: str) -> int: ...


@udf.scalar.builtin
def mismatches(a: str, b: str) -> int: ...


@pytest.mark.parametrize(
    "func",
    [
        hamming,
        jaccard,
        jaro_similarity,
        jaro_winkler_similarity,
        damerau_levenshtein,
        mismatches,
    ],
)
def test_builtin_scalar(con, func):
    a, b = "duck", "luck"
    expr = func(a, b)

    expected = con.raw_sql(f"SELECT {func.__name__}({a!r}, {b!r})").df().squeeze()
    assert con.execute(expr) == expected


def test_builtin_scalar_noargs(con):
    @udf.scalar.builtin
    def version() -> str: ...

    expr = version()
    assert con.execute(expr) == f"v{con.version}"


@udf.agg.builtin
def product(x, where: bool = True) -> float: ...


@udf.agg.builtin
def fsum(x, where: bool = True) -> float: ...


@udf.agg.builtin
def favg(x: float, where: bool = True) -> float: ...


@pytest.mark.parametrize("func", [product, fsum, favg])
def test_builtin_agg(con, func):
    import ibis

    start, stop = 1, 11
    raw_data = list(map(float, range(start, stop)))
    data = ibis.memtable({"a": raw_data})
    expr = func(data.a)

    expected = (
        con.raw_sql(f"SELECT {func.__name__}(a) FROM UNNEST({raw_data!r}) _ (a)")
        .df()
        .squeeze()
    )

    assert con.execute(expr) == expected


@udf.scalar.python
def dont_intercept_null(x: int) -> int:
    assert x is not None
    return x


@pytest.mark.parametrize(
    ("expr", "expected"),
    [
        param(dont_intercept_null(5), 5, id="notnull"),
        param(dont_intercept_null(None), None, id="null"),
        param(dont_intercept_null(5) + dont_intercept_null(None), None, id="mixed"),
    ],
)
def test_dont_intercept_null(con, expr, expected):
    assert con.execute(expr) == expected


def test_kwargs_are_forwarded(con):
    def nullify_two(x: int) -> int:
        return None if x == 2 else x

    @udf.scalar.python
    def no_kwargs(x: int) -> int:
        return nullify_two(x)

    @udf.scalar.python(null_handling="special")
    def with_kwargs(x: int) -> int:
        return nullify_two(x)

    # If we return go Non-NULL -> Non-NULL, then passing null_handling="special"
    # will not change the result
    assert con.execute(no_kwargs(ibis.literal(1))) == 1
    assert con.execute(with_kwargs(ibis.literal(1))) == 1

    # But, if our UDF ever goes Non-NULL -> NULL, then we NEED to pass
    # null_handling="special", otherwise duckdb throws an error
    assert con.execute(with_kwargs(ibis.literal(2))) is None

    expr = no_kwargs(ibis.literal(2))
    with pytest.raises(duckdb.InvalidInputException):
        con.execute(expr)


def test_builtin_udf_uses_dialect():
    # in raw sqlglot, if you call regexp_extract, it will assume the
    # 3rd arg is "position" and not "groups". So when we make the UDF,
    # we need to make sure that we pass the dialect when creating the
    # sqlglot.expressions.func() object
    @udf.scalar.builtin(
        signature=(
            ("string", "string", "array<string>"),
            "struct<y: string, m: string, d: str>",
        ),
    )
    def regexp_extract(s, pattern, groups): ...

    e = regexp_extract("2023-04-15", r"(\d+)-(\d+)-(\d+)", ["y", "m", "d"])
    sql = str(ibis.to_sql(e, dialect="duckdb"))
    assert r"REGEXP_EXTRACT('2023-04-15', '(\d+)-(\d+)-(\d+)', ['y', 'm', 'd'])" in sql



from __future__ import annotations

import polars as pl
import polars.testing
import pytest

import ibis
from ibis.backends.tests.errors import PolarsSQLInterfaceError
from ibis.util import gen_name

pd = pytest.importorskip("pandas")
tm = pytest.importorskip("pandas.testing")


def test_cannot_run_sql_after_drop(con):
    t = con.table("functional_alltypes")
    n = t.count().execute()

    name = gen_name("polars_dot_sql")
    con.create_table(name, t)

    sql = f"SELECT COUNT(*) FROM {name}"

    expr = con.sql(sql)
    result = expr.execute()
    assert result.iat[0, 0] == n

    con.drop_table(name)
    with pytest.raises(PolarsSQLInterfaceError):
        con.sql(sql)


def test_array_flatten(con):
    data = {"id": range(3), "happy": [[["abc"]], [["bcd"]], [["def"]]]}
    t = ibis.memtable(data)
    expr = t.select("id", flat=t.happy.flatten()).order_by("id")
    result = con.to_pyarrow(expr)
    expected = pd.DataFrame(
        {"id": data["id"], "flat": [row[0] for row in data["happy"]]}
    )
    tm.assert_frame_equal(result.to_pandas(), expected)


def test_memtable_polars_types(con):
    # Check that we can create a memtable with some polars-specific types,
    # and that those columns then work in downstream operations
    df = pl.DataFrame(
        {
            "x": ["a", "b", "a"],
            "y": ["c", "d", "c"],
            "z": ["e", "f", "e"],
        },
        schema={
            "x": pl.String,
            "y": pl.Categorical,
            "z": pl.Enum(["e", "f"]),
        },
    )
    t = ibis.memtable(df)
    res = con.to_polars((t.x + t.y + t.z).name("test"))
    sol = (df["x"] + df["y"] + df["z"]).rename("test")
    pl.testing.assert_series_equal(res, sol)


@pytest.mark.parametrize("to_method", ["to_pyarrow", "to_polars"])
def test_streaming(con, mocker, to_method):
    t = con.table("functional_alltypes")
    mocked_collect = mocker.patch("polars.LazyFrame.collect")
    getattr(con, to_method)(t, streaming=True)
    mocked_collect.assert_called_once_with(streaming=True, engine="cpu")


@pytest.mark.parametrize("to_method", ["to_pyarrow", "to_polars"])
def test_engine(con, mocker, to_method):
    t = con.table("functional_alltypes")
    mocked_collect = mocker.patch("polars.LazyFrame.collect")
    getattr(con, to_method)(t, engine="gpu")
    mocked_collect.assert_called_once_with(streaming=False, engine="gpu")



from __future__ import annotations

import pytest

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.types as ir
from ibis import _
from ibis.common.annotations import SignatureValidationError
from ibis.tests.util import assert_pickle_roundtrip


def test_ifelse_method(table):
    bools = table.g.isnull()
    result = bools.ifelse("foo", "bar")
    assert isinstance(result, ir.StringColumn)


def test_ifelse_function_literals():
    res = ibis.ifelse(True, 1, 2)
    sol = ibis.literal(True, type="bool").ifelse(1, 2)
    assert res.equals(sol)

    # condition is explicitly cast to bool
    res = ibis.ifelse(1, 1, 2)
    sol = ibis.literal(1, type="bool").ifelse(1, 2)
    assert res.equals(sol)


def test_ifelse_function_exprs(table):
    res = ibis.ifelse(table.g.isnull(), 1, table.a)
    sol = table.g.isnull().ifelse(1, table.a)
    assert res.equals(sol)

    # condition is cast if not already bool
    res = ibis.ifelse(table.a, 1, table.b)
    sol = table.a.cast("bool").ifelse(1, table.b)
    assert res.equals(sol)


def test_ifelse_function_deferred(table):
    expr = ibis.ifelse(_.g.isnull(), _.a, 2)
    assert repr(expr) == "ifelse(_.g.isnull(), _.a, 2)"
    res = expr.resolve(table)
    sol = table.g.isnull().ifelse(table.a, 2)
    assert res.equals(sol)


def test_case_dshape(table):
    assert isinstance(ibis.cases((True, "bar"), (False, "bar")), ir.Scalar)
    assert isinstance(ibis.cases((True, None), else_="bar"), ir.Scalar)
    assert isinstance(ibis.cases((table.b == 9, None), else_="bar"), ir.Column)
    assert isinstance(ibis.cases((True, table.a), else_=42), ir.Column)
    assert isinstance(ibis.cases((True, 42), else_=table.a), ir.Column)
    assert isinstance(ibis.cases((True, table.a), else_=table.b), ir.Column)

    assert isinstance(ibis.literal(5).cases((9, 42)), ir.Scalar)
    assert isinstance(ibis.literal(5).cases((9, 42), else_=43), ir.Scalar)
    assert isinstance(ibis.literal(5).cases((table.a, 42)), ir.Column)
    assert isinstance(ibis.literal(5).cases((9, table.a)), ir.Column)
    assert isinstance(ibis.literal(5).cases((table.a, table.b)), ir.Column)
    assert isinstance(ibis.literal(5).cases((9, 42), else_=table.a), ir.Column)
    assert isinstance(table.a.cases((9, 42)), ir.Column)
    assert isinstance(table.a.cases((table.b, 42)), ir.Column)
    assert isinstance(table.a.cases((9, table.b)), ir.Column)
    assert isinstance(table.a.cases((table.a, table.b)), ir.Column)


def test_case_dtype():
    assert isinstance(ibis.cases((True, "bar"), (False, "bar")), ir.StringValue)
    assert isinstance(ibis.cases((True, None), else_="bar"), ir.StringValue)
    with pytest.raises(TypeError):
        ibis.cases((True, 5), (False, "bar"))
    with pytest.raises(TypeError):
        ibis.cases((True, 5), else_="bar")


def test_multiple_case_expr(table):
    expr = ibis.cases(
        (table.a == 5, table.f),
        (table.b == 128, table.b * 2),
        (table.c == 1000, table.e),
        else_=table.d,
    )

    # deferred cases
    deferred = ibis.cases(
        (_.a == 5, table.f),
        (_.b == 128, table.b * 2),
        (_.c == 1000, table.e),
        else_=table.d,
    )
    expr2 = deferred.resolve(table)

    # deferred results
    expr3 = ibis.cases(
        (table.a == 5, _.f),
        (table.b == 128, _.b * 2),
        (table.c == 1000, _.e),
        else_=table.d,
    ).resolve(table)

    # deferred default
    expr4 = ibis.cases(
        (table.a == 5, table.f),
        (table.b == 128, table.b * 2),
        (table.c == 1000, table.e),
        else_=_.d,
    ).resolve(table)

    assert expr.equals(expr2)
    assert expr.equals(expr3)
    assert expr.equals(expr4)

    op = expr.op()
    assert isinstance(expr, ir.FloatingColumn)
    assert isinstance(op, ops.SearchedCase)
    assert op.default == table.d.op()


def test_pickle_multiple_case_node(table):
    case1 = table.a == 5
    case2 = table.b == 128
    case3 = table.c == 1000

    result1 = table.f
    result2 = table.b * 2
    result3 = table.e

    default = table.d
    expr = ibis.cases(
        (case1, result1),
        (case2, result2),
        (case3, result3),
        else_=default,
    )

    op = expr.op()
    assert_pickle_roundtrip(op)


def test_simple_case_null_else(table):
    expr = table.g.cases(("foo", "bar"))
    op = expr.op()

    assert isinstance(expr, ir.StringColumn)
    assert isinstance(op.default.to_expr(), ir.Value)


def test_multiple_case_null_else(table):
    expr = ibis.cases((table.g == "foo", "bar"))
    expr2 = ibis.cases((table.g == "foo", _)).resolve("bar")

    assert expr.equals(expr2)

    op = expr.op()
    assert isinstance(expr, ir.StringColumn)
    assert isinstance(op.default.to_expr(), ir.Value)


def test_case_mixed_type():
    t0 = ibis.table(
        [("one", "string"), ("two", "double"), ("three", "int32")],
        name="my_data",
    )

    expr = t0.three.cases((0, "low"), (1, "high"), else_="null").name("label")
    result = t0.select(expr)
    assert result["label"].type().equals(dt.string)


def test_err_on_bad_args(table):
    with pytest.raises(ValueError):
        ibis.cases((True,))
    with pytest.raises(ValueError):
        ibis.cases((True, 3, 4))
    with pytest.raises(ValueError):
        ibis.cases((True, 3, 4))
    with pytest.raises(TypeError):
        ibis.cases((True, 3), 5)


def test_err_on_nonbool_expr(table):
    with pytest.raises(SignatureValidationError):
        ibis.cases((table.a, "bar"), else_="baz")
    with pytest.raises(SignatureValidationError):
        ibis.cases((ibis.literal(1), "bar"), else_=("baz"))


def test_err_on_noncomparable(table):
    table.a.cases((8, "bar"))
    table.a.cases((-8, "bar"))
    # Can't compare an int to a string
    with pytest.raises(TypeError):
        table.a.cases(("foo", "bar"))


def test_err_on_empty_cases(table):
    with pytest.raises(TypeError):
        ibis.cases()
    with pytest.raises(TypeError):
        ibis.cases(else_=42)
    with pytest.raises(TypeError):
        table.a.cases()
    with pytest.raises(TypeError):
        table.a.cases(else_=42)



from __future__ import annotations

import time

import pandas as pd
import pandas.testing as tm
import pytest

import ibis
from ibis import util


@pytest.mark.parametrize(
    "column",
    ["string_col", "double_col", "date_string_col", "timestamp_col"],
)
def test_simple_mv(con, alltypes, column):
    expr = alltypes[[column]].distinct().order_by(column)
    mv_name = util.gen_name("alltypes_mv")
    mv = con.create_materialized_view(mv_name, expr, overwrite=True)
    expected = expr.limit(5).execute()
    result = mv.order_by(column).limit(5).execute()
    tm.assert_frame_equal(result, expected)
    con.drop_materialized_view(mv_name)


def test_mv_on_simple_source(con):
    sc_name = util.gen_name("simple_sc")
    schema = ibis.schema([("v", "int32")])
    # use RisingWave's internal data generator to imitate a upstream data source
    connector_properties = {
        "connector": "datagen",
        "fields.v.kind": "sequence",
        "fields.v.start": "1",
        "fields.v.end": "10",
        "datagen.rows.per.second": "10000",
        "datagen.split.num": "1",
    }
    source = con.create_source(
        sc_name,
        schema,
        connector_properties=connector_properties,
        data_format="PLAIN",
        encode_format="JSON",
    )
    expr = source["v"].sum()
    mv_name = util.gen_name("simple_mv")
    mv = con.create_materialized_view(mv_name, expr)
    # sleep 3s to make sure the data has been generated by the source and consumed by the MV.
    time.sleep(3)
    result = mv.execute()
    expected = pd.DataFrame({"Sum(v)": [55]})
    tm.assert_frame_equal(result, expected)
    con.drop_materialized_view(mv_name)
    con.drop_source(sc_name)


def test_mv_on_table_with_connector(con):
    tblc_name = util.gen_name("simple_table_with_connector")
    schema = ibis.schema([("v", "int32")])
    # use RisingWave's internal data generator to imitate a upstream data source
    connector_properties = {
        "connector": "datagen",
        "fields.v.kind": "sequence",
        "fields.v.start": "1",
        "fields.v.end": "10",
        "datagen.rows.per.second": "10000",
        "datagen.split.num": "1",
    }
    tblc = con.create_table(
        name=tblc_name,
        obj=None,
        schema=schema,
        connector_properties=connector_properties,
        data_format="PLAIN",
        encode_format="JSON",
    )
    expr = tblc["v"].sum()
    mv_name = util.gen_name("simple_mv")
    mv = con.create_materialized_view(mv_name, expr)
    # sleep 1 s to make sure the data has been generated by the source and consumed by the MV.
    time.sleep(1)

    result_tblc = expr.execute()
    assert result_tblc == 55

    result_mv = mv.execute()
    expected = pd.DataFrame({"Sum(v)": [55]})
    tm.assert_frame_equal(result_mv, expected)
    con.drop_materialized_view(mv_name)
    con.drop_table(tblc_name)


def test_sink_from(con, alltypes):
    sk_name = util.gen_name("sk_from")
    connector_properties = {
        "connector": "blackhole",
    }
    con.create_sink(sk_name, "functional_alltypes", connector_properties)
    con.drop_sink(sk_name)


def test_sink_as_select(con, alltypes):
    sk_name = util.gen_name("sk_as_select")
    expr = alltypes[["string_col"]].distinct().order_by("string_col")
    connector_properties = {
        "connector": "blackhole",
    }
    con.create_sink(sk_name, None, connector_properties, obj=expr)
    con.drop_sink(sk_name)



"""Generic value operations."""

from __future__ import annotations

import itertools
from typing import Annotated, Any, Optional
from typing import Literal as LiteralType

from public import public
from typing_extensions import TypeVar

import ibis.expr.datashape as ds
import ibis.expr.datatypes as dt
import ibis.expr.rules as rlz
from ibis.common.annotations import attribute
from ibis.common.deferred import Deferred
from ibis.common.grounds import Singleton
from ibis.common.patterns import InstanceOf, Length
from ibis.common.typing import VarTuple  # noqa: TC001
from ibis.expr.operations.core import Scalar, Unary, Value
from ibis.expr.operations.relations import Relation  # noqa: TC001


@public
class RowID(Value):
    """The row number of the returned result."""

    name = "rowid"
    table: Relation

    shape = ds.columnar
    dtype = dt.int64

    @attribute
    def relations(self):
        return frozenset({self.table})


@public
class Cast(Value):
    """Explicitly cast a value to a specific data type."""

    arg: Value
    to: dt.DataType

    shape = rlz.shape_like("arg")

    @property
    def name(self):
        return f"{self.__class__.__name__}({self.arg.name}, {self.to})"

    @property
    def dtype(self):
        return self.to


@public
class TryCast(Value):
    """Try to cast a value to a specific data type."""

    arg: Value
    to: dt.DataType

    shape = rlz.shape_like("arg")

    @property
    def dtype(self):
        return self.to


@public
class TypeOf(Unary):
    """Return the _database_ data type of the input expression."""

    dtype = dt.string


@public
class IsNull(Unary):
    """Return true if values are null."""

    dtype = dt.boolean


@public
class NotNull(Unary):
    """Returns true if values are not null."""

    dtype = dt.boolean


@public
class NullIf(Value):
    """Return NULL if an expression equals some specific value."""

    arg: Value
    null_if_expr: Value

    dtype = rlz.dtype_like("args")
    shape = rlz.shape_like("args")


@public
class Coalesce(Value):
    """Return the first non-null expression from a tuple of expressions."""

    arg: Annotated[VarTuple[Value], Length(at_least=1)]

    shape = rlz.shape_like("arg")
    dtype = rlz.dtype_like("arg")


@public
class Greatest(Value):
    """Return the largest value from a tuple of expressions."""

    arg: Annotated[VarTuple[Value], Length(at_least=1)]

    shape = rlz.shape_like("arg")
    dtype = rlz.dtype_like("arg")


@public
class Least(Value):
    """Return the smallest value from a tuple of expressions."""

    arg: Annotated[VarTuple[Value], Length(at_least=1)]

    shape = rlz.shape_like("arg")
    dtype = rlz.dtype_like("arg")


T = TypeVar("T", bound=dt.DataType, covariant=True)


@public
class Literal(Scalar[T]):
    """A constant value."""

    value: Annotated[Any, ~InstanceOf(Deferred)]
    dtype: T

    shape = ds.scalar

    def __init__(self, value, dtype):
        # normalize ensures that the value is a valid value for the given dtype
        value = dt.normalize(dtype, value)
        super().__init__(value=value, dtype=dtype)

    @property
    def name(self):
        if self.dtype.is_interval():
            return f"{self.value!r}{self.dtype.unit.short}"
        return repr(self.value)


NULL = Literal(None, dt.null)


@public
class ScalarParameter(Scalar):
    _counter = itertools.count()

    dtype: dt.DataType
    counter: Optional[int] = None

    shape = ds.scalar

    def __init__(self, dtype, counter):
        if counter is None:
            counter = next(self._counter)
        super().__init__(dtype=dtype, counter=counter)

    @property
    def name(self):
        return f"param_{self.counter:d}"


@public
class Constant(Scalar, Singleton):
    """A function that produces a constant."""

    shape = ds.scalar


@public
class Impure(Value):
    pass


@public
class TimestampNow(Constant):
    """Return the current timestamp."""

    dtype = dt.timestamp


@public
class DateNow(Constant):
    """Return the current date."""

    dtype = dt.date


@public
class RandomScalar(Impure):
    """Return a random scalar between 0 and 1."""

    dtype = dt.float64
    shape = ds.scalar


@public
class RandomUUID(Impure):
    """Return a random UUID."""

    dtype = dt.uuid
    shape = ds.scalar


@public
class E(Constant):
    """The mathematical constant e."""

    dtype = dt.float64


@public
class Pi(Constant):
    """The mathematical constant pi."""

    dtype = dt.float64


@public
class Hash(Value):
    """Return the hash of a value."""

    arg: Value

    dtype = dt.int64
    shape = rlz.shape_like("arg")


@public
class HashBytes(Value):
    arg: Value[dt.String | dt.Binary]
    how: LiteralType[
        "md5",
        "MD5",
        "sha1",
        "SHA1",
        "SHA224",
        "sha256",
        "SHA256",
        "sha512",
        "intHash32",
        "intHash64",
        "cityHash64",
        "sipHash64",
        "sipHash128",
    ]

    dtype = dt.binary
    shape = rlz.shape_like("arg")


@public
class HexDigest(Value):
    """Return the hexadecimal digest of a value."""

    arg: Value[dt.String | dt.Binary]
    how: LiteralType[
        "md5",
        "sha1",
        "sha256",
        "sha512",
    ]

    dtype = dt.str
    shape = rlz.shape_like("arg")


# TODO(kszucs): we should merge the case operations by making the
# cases, results and default optional arguments like they are in
# api.py
@public
class SimpleCase(Value):
    """Simple case statement."""

    base: Value
    cases: Annotated[VarTuple[Value], Length(at_least=1)]
    results: Annotated[VarTuple[Value], Length(at_least=1)]
    default: Value

    def __init__(self, base, cases, results, default):
        assert len(cases) == len(results)
        for case in cases:
            if not rlz.comparable(base, case):
                raise TypeError(
                    f"Base expression {rlz.arg_type_error_format(base)} and "
                    f"case {rlz.arg_type_error_format(case)} are not comparable"
                )
        super().__init__(base=base, cases=cases, results=results, default=default)

    @attribute
    def shape(self):
        exprs = [self.base, *self.cases, *self.results, self.default]
        return rlz.highest_precedence_shape(exprs)

    @attribute
    def dtype(self):
        values = [*self.results, self.default]
        return rlz.highest_precedence_dtype(values)


@public
class SearchedCase(Value):
    """Searched case statement."""

    cases: Annotated[VarTuple[Value[dt.Boolean]], Length(at_least=1)]
    results: Annotated[VarTuple[Value], Length(at_least=1)]
    default: Value

    def __init__(self, cases, results, default):
        assert len(cases) == len(results)
        super().__init__(cases=cases, results=results, default=default)

    @attribute
    def shape(self):
        return rlz.highest_precedence_shape((*self.cases, *self.results, self.default))

    @attribute
    def dtype(self):
        exprs = [*self.results, self.default]
        return rlz.highest_precedence_dtype(exprs)


public(NULL=NULL)



from __future__ import annotations  # noqa: INP001

from inspect import signature
from typing import Any

import pytest
from pytest import param

from ibis.backends.tests.signature.typecheck import compatible


def a1(posarg: int): ...


def b1(posarg: str): ...


def a2(posarg: int, **kwargs: Any): ...
def b2(posarg: str, **kwargs): ...


def a3(posarg: int, other_kwarg: bool = True, **kwargs: Any): ...
def b3(posarg: str, **kwargs: Any): ...


def a4(posarg: int, other_kwarg=True, **kwargs: Any): ...
def b4(posarg: str, **kwargs: Any): ...


def a5(posarg: int, /): ...
def b5(posarg2: str, /): ...


@pytest.mark.parametrize(
    "a, b, check_annotations",
    [
        param(
            lambda posarg, *, kwarg1=None, kwarg2=None: ...,
            lambda posarg, *, kwarg2=None, kwarg1=None: ...,
            True,
            id="swapped kwarg order",
        ),
        param(
            lambda posarg, *, kwarg1=None, kwarg2=None, kwarg3=None: ...,
            lambda posarg, *, kwarg2=None, kwarg1=None: ...,
            True,
            id="swapped kwarg order w/extra kwarg first",
        ),
        param(
            lambda posarg, *, kwarg2=None, kwarg1=None: ...,
            lambda posarg, *, kwarg1=None, kwarg2=None, kwarg3=None: ...,
            True,
            id="swapped kwarg order w/extra kwarg second",
        ),
        param(
            a1,
            b1,
            False,
            id="annotations diff types w/out anno check",
        ),
        param(
            a2,
            b3,
            False,
            id="annotations different but parity in annotations",
        ),
        param(
            a3,
            b3,
            False,
            id="annotations different but parity in annotations for matching kwargs",
        ),
        param(
            a4,
            b4,
            False,
            id="annotations different but parity in annotations for matching kwargs",
        ),
        param(
            a2,
            b2,
            False,
            id="annotations different, no anno check, but missing annotation",
        ),
    ],
)
def test_sigs_compatible(a, b, check_annotations):
    sig_a, sig_b = signature(a), signature(b)
    assert compatible(sig_a, sig_b, check_annotations=check_annotations)


@pytest.mark.parametrize(
    "a, b, check_annotations",
    [
        param(
            lambda posarg, /, *, kwarg2=None, kwarg1=None: ...,
            lambda posarg, *, kwarg1=None, kwarg2=None, kwarg3=None: ...,
            True,
            id="one positional only",
        ),
        param(
            lambda posarg, *, kwarg1=None, kwarg2=None: ...,
            lambda posarg, kwarg1=None, kwarg2=None: ...,
            True,
            id="not kwarg only",
        ),
        param(
            a1,
            b1,
            True,
            id="annotations diff types w/anno check",
        ),
        param(
            a2,
            b3,
            True,
            id="annotations different but parity in annotations",
        ),
        param(
            a5,
            b5,
            False,
            id="names different, but positional only",
        ),
    ],
)
def test_sigs_incompatible(a, b, check_annotations):
    sig_a, sig_b = signature(a), signature(b)
    assert not compatible(sig_a, sig_b, check_annotations=check_annotations)



from __future__ import annotations

import sqlite3

import pytest
from packaging.version import parse as vparse
from pytest import param

import ibis
import ibis.common.exceptions as com
import ibis.expr.schema as sch

np = pytest.importorskip("numpy")
pd = pytest.importorskip("pandas")

sqlite_right_or_full_mark = pytest.mark.notyet(
    ["sqlite"],
    condition=vparse(sqlite3.sqlite_version) < vparse("3.39"),
    reason="SQLite < 3.39 doesn't support RIGHT/FULL OUTER joins",
)


def _pandas_semi_join(left, right, on, **_):
    assert len(on) == 1, str(on)
    inner = pd.merge(left, right, how="inner", on=on)
    filt = left.loc[:, on[0]].isin(inner.loc[:, on[0]])
    return left.loc[filt, :]


def _pandas_anti_join(left, right, on, **_):
    inner = pd.merge(left, right, how="left", indicator=True, on=on)
    return inner[inner["_merge"] == "left_only"]


IMPLS = {
    "semi": _pandas_semi_join,
    "anti": _pandas_anti_join,
}


def check_eq(left, right, how, **kwargs):
    impl = IMPLS.get(how, pd.merge)
    return impl(left, right, how=how, **kwargs)


@pytest.mark.parametrize(
    "how",
    [
        "inner",
        "left",
        param(
            "right",
            marks=[sqlite_right_or_full_mark],
        ),
        param(
            "outer",
            # TODO: mysql will likely never support full outer join
            # syntax, but we might be able to work around that using
            # LEFT JOIN UNION RIGHT JOIN
            marks=sqlite_right_or_full_mark,
        ),
    ],
)
@pytest.mark.notimpl(["druid"])
def test_mutating_join(backend, batting, awards_players, how):
    left = batting.filter(batting.yearID == 2015)
    right = awards_players.filter(awards_players.lgID == "NL").drop("yearID", "lgID")

    left_df = left.execute()
    right_df = right.execute()
    predicate = ["playerID"]
    result_order = ["playerID", "yearID", "lgID", "stint"]

    expr = left.join(right, predicate, how=how)
    cols = list(left.columns)
    if how == "inner":
        result = (
            expr.execute()
            .fillna(np.nan)[cols]
            .sort_values(result_order)
            .reset_index(drop=True)
        )
    else:
        result = (
            expr.execute()
            .fillna(np.nan)
            .assign(
                playerID=lambda df: df.playerID.where(
                    df.playerID.notnull(), df.playerID_right
                )
            )
            .drop(["playerID_right"], axis=1)[cols]
            .sort_values(result_order)
            .reset_index(drop=True)
        )

    expected = (
        check_eq(left_df, right_df, how=how, on=predicate, suffixes=("_x", "_y"))[cols]
        .sort_values(result_order)
        .reset_index(drop=True)
    )

    backend.assert_frame_equal(result, expected, check_like=True)


@pytest.mark.parametrize("how", ["semi", "anti"])
@pytest.mark.notimpl(["druid"])
@pytest.mark.notyet(["flink"], reason="Flink doesn't support semi joins or anti joins")
def test_filtering_join(backend, batting, awards_players, how):
    left = batting.filter(batting.yearID == 2015)
    right = awards_players.filter(awards_players.lgID == "NL").drop("yearID", "lgID")

    left_df = left.execute()
    right_df = right.execute()
    predicate = ["playerID"]
    result_order = ["playerID", "yearID", "lgID", "stint"]

    expr = left.join(right, predicate, how=how)
    cols = list(left.columns)
    result = (
        expr.execute()
        .fillna(np.nan)
        .sort_values(result_order)[cols]
        .reset_index(drop=True)
    )

    expected = check_eq(
        left_df, right_df, how=how, on=predicate, suffixes=("", "_y")
    ).sort_values(result_order)[cols]

    backend.assert_frame_equal(result, expected, check_like=True)


def test_join_then_filter_no_column_overlap(awards_players, batting):
    left = batting.filter(batting.yearID == 2015)
    year = left.yearID.name("year")
    left = left.select(year, "RBI")
    right = awards_players.filter(awards_players.lgID == "NL")

    expr = left.join(right, left.year == right.yearID)
    filters = [expr.RBI == 9]
    q = expr.filter(filters)
    assert not q.execute().empty


def test_mutate_then_join_no_column_overlap(batting, awards_players):
    left = batting.mutate(year=batting.yearID).filter(lambda t: t.year == 2015)
    left = left["year", "RBI"]
    right = awards_players
    expr = left.join(right, left.year == right.yearID)
    assert not expr.limit(5).execute().empty


@pytest.mark.notimpl(["druid"])
@pytest.mark.notyet(["flink"], reason="Flink doesn't support semi joins")
@pytest.mark.skip("risingwave")  # TODO(Kexiang): RisingWave's bug, investigating
@pytest.mark.parametrize(
    "func",
    [
        param(lambda left, right: left.semi_join(right, "year"), id="method"),
        param(
            lambda left, right: left.join(right, "year", how="left_semi"),
            id="how_left_semi",
        ),
        param(lambda left, right: left.join(right, "year", how="semi"), id="how_semi"),
    ],
)
def test_semi_join_topk(con, batting, awards_players, func):
    if con.name == "sqlite":
        # TODO: remove after CTE extraction is reimplemented
        pytest.skip("topk -> semi-join performance has increased post SQLGlot refactor")
    elif con.name == "risingwave":
        # e.g., https://github.com/ibis-project/ibis/actions/runs/7900463100/job/21562034052
        pytest.skip("risingwave times out on semi join topk")
    batting = batting.mutate(year=batting.yearID)
    left = func(batting, batting.year.topk(5)).select("year", "RBI")
    expr = left.join(awards_players, left.year == awards_players.yearID)
    assert not expr.limit(5).execute().empty


@pytest.mark.notimpl(["druid", "exasol", "oracle"])
@pytest.mark.notimpl(
    ["postgres", "mssql", "risingwave"],
    raises=com.IbisTypeError,
    reason="postgres can't handle null types columns",
)
def test_join_with_pandas(batting, awards_players):
    batting_filt = batting.filter(lambda t: t.yearID < 1900)
    awards_players_filt = awards_players.filter(lambda t: t.yearID < 1900).execute()
    assert isinstance(awards_players_filt, pd.DataFrame)
    t = ibis.memtable(awards_players_filt)
    expr = batting_filt.join(t, "yearID")
    df = expr.execute()
    assert df.yearID.nunique() == 7


def test_join_with_pandas_non_null_typed_columns(batting, awards_players):
    batting_filt = batting.filter(lambda t: t.yearID < 1900).select("yearID")
    awards_players_filt = (
        awards_players.filter(lambda t: t.yearID < 1900).select("yearID").execute()
    )

    # ensure that none of the columns of either table have type null
    batting_schema = batting_filt.schema()
    assert len(batting_schema) == 1
    assert batting_schema["yearID"].is_integer()

    assert sch.infer(awards_players_filt) == sch.Schema(dict(yearID="int"))
    assert isinstance(awards_players_filt, pd.DataFrame)

    t = ibis.memtable(awards_players_filt)
    expr = batting_filt.join(t, "yearID")
    df = expr.execute()
    assert df.yearID.nunique() == 7


@pytest.mark.parametrize(
    ("predicate", "pandas_value"),
    [
        # Trues
        param(True, True, id="true"),
        param(ibis.literal(True), True, id="true-literal"),
        param([True], True, id="true-list"),
        param([ibis.literal(True)], True, id="true-literal-list"),
        # only trues
        param([True, True], True, id="true-true-list"),
        param(
            [ibis.literal(True), ibis.literal(True)], True, id="true-true-literal-list"
        ),
        param([True, ibis.literal(True)], True, id="true-true-const-expr-list"),
        param([ibis.literal(True), True], True, id="true-true-expr-const-list"),
        # Falses
        param(False, False, id="false"),
        param(ibis.literal(False), False, id="false-literal"),
        param([False], False, id="false-list"),
        param([ibis.literal(False)], False, id="false-literal-list"),
        # only falses
        param([False, False], False, id="false-false-list"),
        param(
            [ibis.literal(False), ibis.literal(False)],
            False,
            id="false-false-literal-list",
        ),
        param([False, ibis.literal(False)], False, id="false-false-const-expr-list"),
        param([ibis.literal(False), False], False, id="false-false-expr-const-list"),
    ],
)
@pytest.mark.parametrize(
    "how",
    [
        "inner",
        "left",
        param("right", marks=[sqlite_right_or_full_mark]),
        param("outer", marks=[sqlite_right_or_full_mark]),
    ],
)
def test_join_with_trivial_predicate(awards_players, predicate, how, pandas_value):
    n = 5

    base = awards_players.limit(n)

    left = base.select(left_key="playerID")
    right = base.select(right_key="playerID")

    left_df = pd.DataFrame({"key": [True] * n})
    right_df = pd.DataFrame({"key": [pandas_value] * n})

    expected = pd.merge(left_df, right_df, on="key", how=how)

    expr = left.join(right, predicate, how=how)
    result = expr.to_pandas()

    assert len(result) == len(expected)


@pytest.mark.notimpl(["druid"], raises=com.TableNotFound)
@pytest.mark.parametrize(
    ("how", "nrows", "gen_right", "keys"),
    [
        param(
            "left",
            2,
            lambda left: left.filter(lambda t: t.x == 1).select(y=lambda t: t.x),
            [("x", "y")],
            id="left-xy",
        ),
        param(
            "left",
            2,
            lambda left: left.filter(lambda t: t.x == 1),
            "x",
            id="left-x",
        ),
        param(
            "right",
            1,
            lambda left: left.filter(lambda t: t.x == 1).select(y=lambda t: t.x),
            [("x", "y")],
            id="right-xy",
            marks=[sqlite_right_or_full_mark],
        ),
        param(
            "right",
            1,
            lambda left: left.filter(lambda t: t.x == 1),
            "x",
            id="right-x",
            marks=[sqlite_right_or_full_mark],
        ),
        param(
            "outer",
            2,
            lambda left: left.filter(lambda t: t.x == 1).select(y=lambda t: t.x),
            [("x", "y")],
            id="outer-xy",
            marks=[sqlite_right_or_full_mark],
        ),
        param(
            "outer",
            2,
            lambda left: left.filter(lambda t: t.x == 1),
            "x",
            id="outer-x",
            marks=[sqlite_right_or_full_mark],
        ),
    ],
)
def test_outer_join_nullability(backend, how, nrows, gen_right, keys):
    win = backend.win
    left = win.select(x=lambda t: t.x.cast(t.x.type().copy(nullable=False))).filter(
        lambda t: t.x.isin((1, 2))
    )
    right = gen_right(left)
    expr = left.join(right, keys, how=how)
    assert all(typ.nullable for typ in expr.schema().types)

    result = expr.to_pyarrow()
    assert len(result) == nrows


def test_complex_join_agg(snapshot):
    t1 = ibis.table(dict(value1="float", key1="string", key2="string"), name="table1")
    t2 = ibis.table(dict(value2="float", key1="string", key4="string"), name="table2")

    avg_diff = (t1.value1 - t2.value2).mean()
    expr = t1.left_join(t2, "key1").group_by(t1.key1).aggregate(avg_diff=avg_diff)

    snapshot.assert_match(str(ibis.to_sql(expr, dialect="duckdb")), "out.sql")


def test_join_conflicting_columns(backend, con):
    # GH #7345
    t1 = ibis.memtable({"x": [1, 2, 3], "y": [4, 5, 6], "z": ["a", "b", "c"]})
    t2 = ibis.memtable({"x": [3, 2, 1], "y": [7, 8, 9], "z": ["d", "e", "f"]})

    expr = t1.join(t2, "x")
    result = con.execute(expr).sort_values("x")

    expected = pd.DataFrame(
        {
            "x": [1, 2, 3],
            "y": [4, 5, 6],
            "z": ["a", "b", "c"],
            "y_right": [9, 8, 7],
            "z_right": ["f", "e", "d"],
        }
    )
    backend.assert_frame_equal(result, expected)


@pytest.mark.never(
    [
        "bigquery",
        "clickhouse",
        "datafusion",
        "druid",
        "exasol",
        "flink",
        "impala",
        "mssql",
        "mysql",
        "oracle",
        "postgres",
        "pyspark",
        "risingwave",
        "snowflake",
        "sqlite",
        "trino",
        "databricks",
    ],
    reason="Users can implement this with ibis.row_number(): https://github.com/ibis-project/ibis/issues/9486",
)
def test_positional_join(backend, con):
    t1 = ibis.memtable({"x": [1, 2, 3]})
    t2 = ibis.memtable({"x": [3, 2, 1]})
    j = t1.join(t2, how="positional")
    result = con.execute(j)
    expected = pd.DataFrame({"x": [1, 2, 3], "x_right": [3, 2, 1]})
    backend.assert_frame_equal(result, expected)



from __future__ import annotations

import pytest
from pytest import param

import ibis
import ibis.common.exceptions as exc
import ibis.expr.datatypes as dt
from ibis.backends.tests.errors import PsycoPg2InternalError, Py4JJavaError

np = pytest.importorskip("numpy")
pd = pytest.importorskip("pandas")
tm = pytest.importorskip("pandas.testing")
pa = pytest.importorskip("pyarrow")

pytestmark = [
    pytest.mark.never(
        ["sqlite", "mysql", "mssql"], reason="Unlikely to ever add map support"
    ),
    pytest.mark.notyet(
        ["bigquery", "impala"], reason="Backend doesn't yet implement map types"
    ),
    pytest.mark.notimpl(
        ["exasol", "polars", "druid", "oracle"],
        reason="Not yet implemented in ibis",
    ),
]

mark_notyet_postgres = pytest.mark.notyet(
    "postgres", reason="only supports string -> string"
)

mark_notyet_snowflake = pytest.mark.notyet(
    "snowflake", reason="map keys must be strings"
)

mark_notimpl_risingwave_hstore = pytest.mark.notimpl(
    ["risingwave"],
    reason="function hstore(character varying[], character varying[]) does not exist",
)

mark_notyet_datafusion = pytest.mark.notyet(
    ["datafusion"], raises=Exception, reason="only map and make_map are available"
)


@pytest.mark.notyet("clickhouse", reason="nested types can't be NULL")
@pytest.mark.notimpl(
    ["risingwave"],
    raises=PsycoPg2InternalError,
    reason="function hstore(character varying[], character varying[]) does not exist",
)
@pytest.mark.parametrize(
    ("k", "v"),
    [
        param(["a", "b"], None, id="null_values"),
        param(None, ["c", "d"], id="null_keys"),
        param(None, None, id="null_both"),
    ],
)
@mark_notyet_datafusion
def test_map_nulls(con, k, v):
    k = ibis.literal(k, type="array<string>")
    v = ibis.literal(v, type="array<string>")
    m = ibis.map(k, v)
    assert con.execute(m) is None


@pytest.mark.notyet("clickhouse", reason="nested types can't be NULL")
@pytest.mark.notimpl(
    ["risingwave"],
    raises=PsycoPg2InternalError,
    reason="function hstore(character varying[], character varying[]) does not exist",
)
@pytest.mark.parametrize(
    ("k", "v"),
    [
        param(None, ["c", "d"], id="null_keys"),
        param(None, None, id="null_both"),
    ],
)
@mark_notyet_datafusion
def test_map_keys_nulls(con, k, v):
    k = ibis.literal(k, type="array<string>")
    v = ibis.literal(v, type="array<string>")
    m = ibis.map(k, v)
    assert con.execute(m.keys()) is None


@pytest.mark.notyet("clickhouse", reason="nested types can't be NULL")
@pytest.mark.notimpl(
    ["risingwave"],
    raises=PsycoPg2InternalError,
    reason="function hstore(character varying[], character varying[]) does not exist",
)
@pytest.mark.parametrize(
    "map",
    [
        param(
            ibis.map(
                ibis.literal(["a", "b"]), ibis.literal(None, type="array<string>")
            ),
            id="null_values",
        ),
        param(
            ibis.map(
                ibis.literal(None, type="array<string>"),
                ibis.literal(None, type="array<string>"),
            ),
            id="null_both",
        ),
        param(ibis.literal(None, type="map<string, string>"), id="null_map"),
    ],
)
@mark_notyet_datafusion
def test_map_values_nulls(con, map):
    assert con.execute(map.values()) is None


@pytest.mark.notimpl(
    ["risingwave"],
    raises=PsycoPg2InternalError,
    reason="function hstore(character varying[], character varying[]) does not exist",
)
@pytest.mark.parametrize(
    ("map", "key"),
    [
        param(
            ibis.map(
                ibis.literal(["a", "b"]), ibis.literal(["c", "d"], type="array<string>")
            ),
            ibis.literal(None, type="string"),
            marks=[
                pytest.mark.notimpl(
                    "flink",
                    raises=AssertionError,
                    reason="not yet implemented",
                    strict=False,
                ),
            ],
            id="non_null_map_null_key",
        ),
        param(
            ibis.map(
                ibis.literal(None, type="array<string>"),
                ibis.literal(None, type="array<string>"),
            ),
            "a",
            marks=[
                pytest.mark.notyet("clickhouse", reason="nested types can't be NULL")
            ],
            id="null_both_non_null_key",
        ),
        param(
            ibis.map(
                ibis.literal(None, type="array<string>"),
                ibis.literal(None, type="array<string>"),
            ),
            ibis.literal(None, type="string"),
            marks=[
                pytest.mark.notyet("clickhouse", reason="nested types can't be NULL"),
            ],
            id="null_both_null_key",
        ),
        param(
            ibis.literal(None, type="map<string, string>"),
            "a",
            marks=[
                pytest.mark.notyet("clickhouse", reason="nested types can't be NULL")
            ],
            id="null_map_non_null_key",
        ),
        param(
            ibis.literal(None, type="map<string, string>"),
            ibis.literal(None, type="string"),
            marks=[
                pytest.mark.notyet("clickhouse", reason="nested types can't be NULL")
            ],
            id="null_map_null_key",
        ),
    ],
)
@pytest.mark.parametrize("method", ["get", "contains"])
@mark_notyet_datafusion
def test_map_get_contains_nulls(con, map, key, method):
    expr = getattr(map, method)
    assert con.execute(expr(key)) is None


@pytest.mark.notyet("clickhouse", reason="nested types can't be NULL")
@pytest.mark.notimpl(
    ["risingwave"],
    raises=PsycoPg2InternalError,
    reason="function hstore(character varying[], character varying[]) does not exist",
)
@pytest.mark.parametrize(
    ("m1", "m2"),
    [
        param(
            ibis.literal(None, type="map<string, string>"),
            ibis.literal({"a": "b"}, type="map<string, string>"),
            id="null_and_non_null",
        ),
        param(
            ibis.literal({"a": "b"}, type="map<string, string>"),
            ibis.literal(None, type="map<string, string>"),
            id="non_null_and_null",
        ),
        param(
            ibis.literal(None, type="map<string, string>"),
            ibis.literal(None, type="map<string, string>"),
            id="null_and_null",
        ),
    ],
)
@mark_notyet_datafusion
def test_map_merge_nulls(con, m1, m2):
    concatted = m1 + m2
    assert con.execute(concatted) is None


@mark_notyet_datafusion
def test_map_table(backend):
    table = backend.map
    assert table.kv.type().is_map()
    assert not table.limit(1).execute().empty


@mark_notimpl_risingwave_hstore
@mark_notyet_datafusion
def test_column_map_values(backend):
    table = backend.map
    expr = table.select("idx", vals=table.kv.values()).order_by("idx")
    result = expr.execute().vals
    expected = pd.Series([[1, 2, 3], [4, 5, 6]], name="vals")
    backend.assert_series_equal(result, expected)


@mark_notyet_datafusion
@pytest.mark.notyet(
    ["databricks"], reason="says one thing, does something completely different"
)
def test_column_map_merge(backend):
    table = backend.map
    expr = table.select(
        "idx",
        merged=table.kv + ibis.map({"d": np.int64(1)}),
    ).order_by("idx")
    result = expr.execute().merged
    expected = pd.Series(
        [{"a": 1, "b": 2, "c": 3, "d": 1}, {"d": 1, "e": 5, "f": 6}], name="merged"
    )
    tm.assert_series_equal(result, expected)


@mark_notimpl_risingwave_hstore
@mark_notyet_datafusion
def test_literal_map_keys(con):
    mapping = ibis.literal({"1": "a", "2": "b"})
    expr = mapping.keys().name("tmp")

    result = con.execute(expr)
    # This does not check whether `result` is an np.array or a list,
    # because it varies across backends and backend configurations
    assert np.array_equal(result, ["1", "2"])


@mark_notimpl_risingwave_hstore
@mark_notyet_datafusion
def test_literal_map_values(con):
    mapping = ibis.literal({"1": "a", "2": "b"})
    expr = mapping.values().name("tmp")

    result = con.execute(expr)
    assert np.array_equal(result, ["a", "b"])


@mark_notimpl_risingwave_hstore
@mark_notyet_postgres
@mark_notyet_datafusion
def test_scalar_isin_literal_map_keys(con):
    mapping = ibis.literal({"a": 1, "b": 2})
    a = ibis.literal("a")
    c = ibis.literal("c")
    true = a.isin(mapping.keys())
    false = c.isin(mapping.keys())
    assert con.execute(true) == True  # noqa: E712
    assert con.execute(false) == False  # noqa: E712


@mark_notimpl_risingwave_hstore
@mark_notyet_postgres
@mark_notyet_datafusion
def test_map_scalar_contains_key_scalar(con):
    mapping = ibis.literal({"a": 1, "b": 2})
    a = ibis.literal("a")
    c = ibis.literal("c")
    true = mapping.contains(a)
    false = mapping.contains(c)
    assert con.execute(true) == True  # noqa: E712
    assert con.execute(false) == False  # noqa: E712


@mark_notimpl_risingwave_hstore
@mark_notyet_datafusion
def test_map_scalar_contains_key_column(backend, alltypes, df):
    value = {"1": "a", "3": "c"}
    mapping = ibis.literal(value)
    expr = mapping.contains(alltypes.string_col).name("tmp")
    result = expr.execute()
    expected = df.string_col.apply(lambda x: x in value).rename("tmp")
    backend.assert_series_equal(result, expected)


@mark_notimpl_risingwave_hstore
@mark_notyet_postgres
@mark_notyet_datafusion
def test_map_column_contains_key_scalar(backend, alltypes, df):
    expr = ibis.map(ibis.array([alltypes.string_col]), ibis.array([alltypes.int_col]))
    series = df.apply(lambda row: {row["string_col"]: row["int_col"]}, axis=1)

    result = expr.contains("1").name("tmp").execute()
    series = series.apply(lambda x: "1" in x).rename("tmp")

    backend.assert_series_equal(result, series)


@mark_notimpl_risingwave_hstore
@mark_notyet_postgres
@mark_notyet_datafusion
def test_map_column_contains_key_column(alltypes):
    map_expr = ibis.map(
        ibis.array([alltypes.string_col]), ibis.array([alltypes.int_col])
    )
    expr = map_expr.contains(alltypes.string_col).name("tmp")
    result = expr.execute()
    assert result.all()


@mark_notimpl_risingwave_hstore
@mark_notyet_postgres
@mark_notyet_datafusion
@pytest.mark.notyet(
    ["databricks"], reason="says one thing, does something completely different"
)
def test_literal_map_merge(con):
    a = ibis.literal({"a": 0, "b": 2})
    b = ibis.literal({"a": 1, "c": 3})
    expr = a + b

    assert con.execute(expr) == {"a": 1, "b": 2, "c": 3}


@mark_notimpl_risingwave_hstore
@mark_notyet_datafusion
def test_literal_map_getitem_broadcast(backend, alltypes, df):
    value = {"1": "a", "2": "b"}

    lookup_table = ibis.literal(value)
    expr = lookup_table[alltypes.string_col]

    result = expr.name("tmp").execute()
    expected = df.string_col.apply(value.get).rename("tmp")

    backend.assert_series_equal(result, expected)


keys = pytest.mark.parametrize(
    "keys",
    [
        pytest.param(["a", "b"], id="string"),
        pytest.param(
            [1, 2],
            marks=[mark_notyet_postgres, mark_notyet_snowflake],
            id="int",
        ),
        pytest.param(
            [True, False],
            marks=[mark_notyet_postgres, mark_notyet_snowflake],
            id="bool",
        ),
        pytest.param(
            [1.0, 2.0],
            marks=[
                pytest.mark.notyet(
                    "clickhouse",
                    reason="only supports str,int,bool,timestamp keys",
                    strict=False,
                ),
                mark_notyet_postgres,
                mark_notyet_snowflake,
            ],
            id="float",
        ),
        pytest.param(
            [ibis.timestamp("2021-01-01"), ibis.timestamp("2021-01-02")],
            marks=[mark_notyet_postgres, mark_notyet_snowflake],
            id="timestamp",
        ),
        pytest.param(
            [ibis.date(1, 2, 3), ibis.date(4, 5, 6)],
            marks=[
                pytest.mark.notyet(
                    "clickhouse", reason="only supports str,int,bool,timestamp keys"
                ),
                mark_notyet_postgres,
                mark_notyet_snowflake,
            ],
            id="date",
        ),
        pytest.param(
            [[1, 2], [3, 4]],
            marks=[
                pytest.mark.notyet(
                    "clickhouse", reason="only supports str,int,bool,timestamp keys"
                ),
                mark_notyet_postgres,
                mark_notyet_snowflake,
            ],
            id="array",
        ),
        pytest.param(
            [ibis.struct(dict(a=1)), ibis.struct(dict(a=2))],
            marks=[
                pytest.mark.notyet(
                    "clickhouse", reason="only supports str,int,bool,timestamp keys"
                ),
                mark_notyet_postgres,
                pytest.mark.notyet(
                    ["flink"],
                    raises=Py4JJavaError,
                    reason="does not support selecting struct key from map",
                ),
                mark_notyet_snowflake,
            ],
            id="struct",
        ),
    ],
)


values = pytest.mark.parametrize(
    "values",
    [
        pytest.param(["a", "b"], id="string"),
        pytest.param(
            [1, 2],
            marks=[
                mark_notyet_postgres,
            ],
            id="int",
        ),
        pytest.param(
            [True, False],
            marks=[
                mark_notyet_postgres,
            ],
            id="bool",
        ),
        pytest.param(
            [1.0, 2.0],
            marks=[
                mark_notyet_postgres,
            ],
            id="float",
        ),
        pytest.param(
            [ibis.timestamp("2021-01-01"), ibis.timestamp("2021-01-02")],
            marks=[
                mark_notyet_postgres,
            ],
            id="timestamp",
        ),
        pytest.param(
            [ibis.date(2021, 1, 1), ibis.date(2022, 2, 2)],
            marks=[mark_notyet_postgres],
            id="date",
        ),
        pytest.param(
            [[1, 2], [3, 4]],
            marks=[
                pytest.mark.notyet("clickhouse", reason="nested types can't be null"),
                mark_notyet_postgres,
            ],
            id="array",
        ),
        pytest.param(
            [ibis.struct(dict(a=1)), ibis.struct(dict(a=2))],
            marks=[
                pytest.mark.notyet("clickhouse", reason="nested types can't be null"),
                mark_notyet_postgres,
            ],
            id="struct",
        ),
    ],
)


@values
@keys
@mark_notimpl_risingwave_hstore
@mark_notyet_datafusion
def test_map_get_all_types(con, keys, values):
    m = ibis.map(ibis.array(keys), ibis.array(values))
    for key, val in zip(keys, values):
        if isinstance(val, ibis.Expr):
            val = con.execute(val)
        assert con.execute(m[key]) == val


@keys
@mark_notimpl_risingwave_hstore
@mark_notyet_datafusion
def test_map_contains_all_types(con, keys):
    a = ibis.array(keys)
    m = ibis.map(a, a)
    for key in keys:
        assert con.execute(m.contains(key))


@mark_notimpl_risingwave_hstore
@mark_notyet_datafusion
def test_literal_map_get_broadcast(backend, alltypes, df):
    value = {"1": "a", "2": "b"}

    lookup_table = ibis.literal(value)
    expr = lookup_table.get(alltypes.string_col, "default")

    result = expr.name("tmp").execute()
    expected = df.string_col.apply(lambda x: value.get(x, "default")).rename("tmp")

    backend.assert_series_equal(result, expected)


@pytest.mark.parametrize(
    ("keys", "values"),
    [
        param(
            ["a", "b"],
            [1, 2],
            id="string",
            marks=pytest.mark.notyet(
                ["postgres", "risingwave"],
                reason="only support maps of string -> string",
            ),
        ),
        param(["a", "b"], ["1", "2"], id="int"),
    ],
)
@mark_notimpl_risingwave_hstore
def test_map_construct_dict(con, keys, values):
    expr = ibis.map(keys, values)
    result = con.execute(expr.name("tmp"))
    assert result == dict(zip(keys, values))


@pytest.mark.notimpl(
    ["flink"],
    raises=pa.lib.ArrowInvalid,
    reason="Map array child array should have no nulls",
)
@mark_notimpl_risingwave_hstore
@mark_notyet_postgres
def test_map_construct_array_column(con, alltypes, df):
    expr = ibis.map(ibis.array([alltypes.string_col]), ibis.array([alltypes.int_col]))
    result = con.execute(expr)
    expected = df.apply(lambda row: {row["string_col"]: row["int_col"]}, axis=1)

    assert result.to_list() == expected.to_list()


@mark_notimpl_risingwave_hstore
@mark_notyet_postgres
@mark_notyet_datafusion
def test_map_get_with_compatible_value_smaller(con):
    value = ibis.literal({"A": 1000, "B": 2000})
    expr = value.get("C", 3)
    assert con.execute(expr) == 3


@mark_notimpl_risingwave_hstore
@mark_notyet_postgres
@mark_notyet_datafusion
def test_map_get_with_compatible_value_bigger(con):
    value = ibis.literal({"A": 1, "B": 2})
    expr = value.get("C", 3000)
    assert con.execute(expr) == 3000


@mark_notimpl_risingwave_hstore
@mark_notyet_postgres
@mark_notyet_datafusion
def test_map_get_with_incompatible_value_different_kind(con):
    value = ibis.literal({"A": 1000, "B": 2000})
    expr = value.get("C", 3.0)
    assert con.execute(expr) == 3.0


@mark_notimpl_risingwave_hstore
@mark_notyet_postgres
@mark_notyet_datafusion
@pytest.mark.parametrize("null_value", [None, ibis.null()])
def test_map_get_with_null_on_not_nullable(con, null_value):
    map_type = dt.Map(dt.string, dt.Int16(nullable=False))
    value = ibis.literal({"A": 1000, "B": 2000}).cast(map_type)
    expr = value.get("C", null_value)
    result = con.execute(expr)
    assert pd.isna(result)


@pytest.mark.parametrize("null_value", [None, ibis.null()])
@pytest.mark.notyet(
    ["flink"], raises=Py4JJavaError, reason="Flink cannot handle typeless nulls"
)
@mark_notimpl_risingwave_hstore
@mark_notyet_datafusion
def test_map_get_with_null_on_null_type_with_null(con, null_value):
    value = ibis.literal({"A": None, "B": None})
    expr = value.get("C", null_value)
    result = con.execute(expr)
    assert pd.isna(result)


@pytest.mark.notyet(
    ["flink"], raises=Py4JJavaError, reason="Flink cannot handle typeless nulls"
)
@mark_notimpl_risingwave_hstore
@mark_notyet_postgres
@mark_notyet_datafusion
def test_map_get_with_null_on_null_type_with_non_null(con):
    value = ibis.literal({"A": None, "B": None})
    expr = value.get("C", 1)
    assert con.execute(expr) == 1


@pytest.mark.notimpl(
    ["flink"],
    raises=exc.IbisError,
    reason="`tbl_properties` is required when creating table with schema",
)
@mark_notimpl_risingwave_hstore
@mark_notyet_datafusion
def test_map_create_table(con, temp_table):
    t = con.create_table(
        temp_table,
        schema=ibis.schema(dict(xyz="map<string, string>")),
    )
    assert t.schema()["xyz"].is_map()


@pytest.mark.notimpl(
    ["flink"],
    raises=exc.OperationNotDefinedError,
    reason="No translation rule for <class 'ibis.expr.operations.maps.MapLength'>",
)
@mark_notimpl_risingwave_hstore
@mark_notyet_datafusion
def test_map_length(con):
    expr = ibis.literal(dict(a="A", b="B")).length()
    assert con.execute(expr) == 2


@mark_notyet_datafusion
def test_map_keys_unnest(backend):
    expr = backend.map.kv.keys().unnest()
    result = expr.to_pandas()
    assert frozenset(result) == frozenset("abcdef")


@mark_notimpl_risingwave_hstore
@mark_notyet_datafusion
def test_map_contains_null(con):
    expr = ibis.map(["a"], ibis.literal([None], type="array<string>"))
    assert con.execute(expr.contains("a"))
    assert not con.execute(expr.contains("b"))



from __future__ import annotations

import pytest
from pytest import param

import ibis
import ibis.expr.datatypes as dt

pl = pytest.importorskip("polars")

from ibis.formats.polars import PolarsData, PolarsSchema, PolarsType  # noqa: E402


@pytest.mark.parametrize(
    ("ibis_dtype", "polars_type"),
    [
        param(dt.bool, pl.Boolean, id="bool"),
        param(dt.null, pl.Null, id="null"),
        param(dt.Array(dt.string), pl.List(pl.Utf8), id="array_string"),
        param(dt.string, pl.Utf8, id="string"),
        param(dt.binary, pl.Binary, id="binary"),
        param(dt.date, pl.Date, id="date"),
        param(dt.time, pl.Time, id="time"),
        param(dt.int8, pl.Int8, id="int8"),
        param(dt.int16, pl.Int16, id="int16"),
        param(dt.int32, pl.Int32, id="int32"),
        param(dt.int64, pl.Int64, id="int64"),
        param(dt.uint8, pl.UInt8, id="uint8"),
        param(dt.uint16, pl.UInt16, id="uint16"),
        param(dt.uint32, pl.UInt32, id="uint32"),
        param(dt.uint64, pl.UInt64, id="uint64"),
        param(dt.float32, pl.Float32, id="float32"),
        param(dt.float64, pl.Float64, id="float64"),
        param(
            dt.Timestamp(scale=9), pl.Datetime("ns", time_zone=None), id="timestamp_ns"
        ),
        param(
            dt.Timestamp(scale=6), pl.Datetime("us", time_zone=None), id="timestamp_us"
        ),
        param(
            dt.Timestamp(scale=3), pl.Datetime("ms", time_zone=None), id="timestamp_ms"
        ),
        param(
            dt.Timestamp("UTC", scale=9),
            pl.Datetime("ns", time_zone="UTC"),
            id="timestamp_tz",
        ),
        param(dt.Interval(unit="ms"), pl.Duration("ms"), id="interval_ms"),
        param(dt.Interval(unit="us"), pl.Duration("us"), id="interval_us"),
        param(dt.Interval(unit="ns"), pl.Duration("ns"), id="interval_ns"),
        param(
            dt.Struct(
                dict(a=dt.string, b=dt.Array(dt.Array(dt.Struct(dict(c=dt.float64)))))
            ),
            pl.Struct(
                [
                    pl.Field("a", pl.Utf8),
                    pl.Field(
                        "b", pl.List(pl.List(pl.Struct([pl.Field("c", pl.Float64)])))
                    ),
                ]
            ),
            id="struct",
        ),
    ],
)
def test_to_from_ibis_type(ibis_dtype, polars_type):
    assert PolarsType.from_ibis(ibis_dtype) == polars_type
    assert PolarsType.to_ibis(polars_type) == ibis_dtype
    assert PolarsType.to_ibis(polars_type, nullable=False) == ibis_dtype(nullable=False)


@pytest.mark.parametrize(
    ("ibis_dtype", "polars_type"),
    [
        param(
            dt.Timestamp(scale=0), pl.Datetime("ns", time_zone=None), id="timestamp_sc0"
        ),
        param(dt.Timestamp(), pl.Datetime("ns", time_zone=None), id="timestamp_scn"),
        param(
            dt.Timestamp("UTC", scale=0),
            pl.Datetime("ns", time_zone="UTC"),
            id="timestamp_tz_sc0",
        ),
    ],
)
def test_from_ibis_type_seconds(ibis_dtype, polars_type):
    # we accept seconds as an ibis type, default to ns in polars
    assert PolarsType.from_ibis(ibis_dtype) == polars_type


def test_decimal():
    assert PolarsType.to_ibis(pl.Decimal()) == dt.Decimal(precision=None, scale=0)
    assert PolarsType.to_ibis(pl.Decimal(precision=6, scale=3)) == dt.Decimal(
        precision=6, scale=3
    )
    assert PolarsType.from_ibis(dt.Decimal()) == pl.Decimal(precision=None, scale=9)
    assert PolarsType.from_ibis(dt.Decimal(precision=6, scale=3)) == pl.Decimal(
        precision=6, scale=3
    )


def test_enum_categorical():
    assert PolarsType.to_ibis(pl.Categorical()) == dt.string
    assert PolarsType.to_ibis(pl.Enum(["a", "b"])) == dt.string


def test_interval_unsupported_unit():
    typ = dt.Interval(unit="s")
    with pytest.raises(ValueError, match="Unsupported polars duration unit"):
        PolarsType.from_ibis(typ)


def test_map_unsupported():
    typ = dt.Map(dt.String(), dt.Int64())
    with pytest.raises(NotImplementedError, match="to polars is not supported"):
        PolarsType.from_ibis(typ)


def test_schema_to_and_from_ibis():
    polars_schema = {"x": pl.Int64, "y": pl.List(pl.Utf8)}
    ibis_schema = ibis.schema({"x": "int64", "y": "array<string>"})

    s1 = PolarsSchema.to_ibis(polars_schema)
    assert s1.equals(ibis_schema)

    s2 = PolarsSchema.from_ibis(ibis_schema)
    assert s2 == polars_schema


def test_infer_scalar():
    assert PolarsData.infer_scalar(1).is_integer()
    nested = PolarsData.infer_scalar([1])
    assert nested.is_array()
    assert nested.value_type.is_integer()


def test_infer_column():
    assert PolarsData.infer_column([1, 2, None]).is_integer()
    assert PolarsData.infer_column(["a", "b"]).is_string()


def test_infer_table():
    schema = PolarsData.infer_table({"x": [1, 2, None], "y": ["a", "b", "c"]})
    assert schema.names == ("x", "y")
    assert schema["x"].is_integer()
    assert schema["y"].is_string()


def test_convert_scalar():
    df = pl.DataFrame({"x": ["1"]})
    res = PolarsData.convert_scalar(df, dt.int64)
    assert res == 1
    assert isinstance(res, int)


def test_convert_column():
    df = pl.DataFrame({"x": ["1", "2"]})
    res = PolarsData.convert_column(df, dt.int64)
    sol = pl.Series(name="x", values=[1, 2], dtype=pl.Int64)
    assert res.equals(sol)
    assert res.dtype == sol.dtype


def test_convert_table():
    df = pl.DataFrame({"x": ["1", "2"], "z": ["a", "b"]})
    schema = ibis.schema({"x": "int64", "z": "string"})
    df = PolarsData.convert_table(df, schema)
    sol = pl.DataFrame(
        {"x": [1, 2], "z": ["a", "b"]}, schema={"x": pl.Int64, "z": pl.Utf8}
    )
    assert df.equals(sol)
    assert df.schema == sol.schema


def test_array_type():
    assert PolarsType.to_ibis(pl.Array(pl.Int64, 2)) == dt.Array(dt.int64)
